/*
 * COPYRIGHT:		 See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System for NT
 * FILE:             Ext2fs.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:		 http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ntifs.h"
#include "ext2fs.h"

/* GLOBALS ***************************************************************/

PEXT2_GLOBAL	Ext2Global = NULL;

/* FUNCTIONS ***************************************************************/

LARGE_INTEGER
Ext2SysTime (IN ULONG i_time)
{
	LARGE_INTEGER SysTime;
	SysTime.QuadPart = (LONGLONG)(i_time) * 10000000;
	SysTime.QuadPart += Ext2Global->TimeZone.QuadPart;
	return SysTime;
}

ULONG
Ext2InodeTime (IN LARGE_INTEGER SysTime)
{
	return (ULONG)((SysTime.QuadPart - Ext2Global->TimeZone.QuadPart) / 10000000);
}


NTSTATUS
Ext2ReadSectors(IN PDEVICE_OBJECT pDeviceObject,
		IN ULONG DiskSector,
		IN ULONG SectorCount,
		IN PVOID Buffer)
{
	LARGE_INTEGER   sectorNumber;
	PIRP            irp;
	IO_STATUS_BLOCK ioStatus;
	KEVENT          event;
	NTSTATUS        Status;
	ULONG           sectorSize;
	
//	KdPrint(("Ext2ReadSectors(pDeviceObject %xh, lba %xh, sectors: %xh Buffer %xh)\n",
//		pDeviceObject,DiskSector, SectorCount, Buffer));
	
	sectorNumber.QuadPart = (LONGLONG)(DiskSector) * SECTOR_SIZE;
	
	KeInitializeEvent(&event, NotificationEvent, FALSE);
	
	sectorSize = SECTOR_SIZE * SectorCount;
	
	irp = IoBuildSynchronousFsdRequest(IRP_MJ_READ,
		pDeviceObject,
		Buffer,
		sectorSize,
		&sectorNumber,
		&event,
		&ioStatus );
	
	if (irp == NULL) 
	{
		return STATUS_UNSUCCESSFUL;
	}
	
	Status = IoCallDriver(pDeviceObject, irp);
	
	if (Status == STATUS_PENDING) 
	{
		KeWaitForSingleObject(&event,
			Suspended,
			KernelMode,
			FALSE,
			NULL);
		Status = ioStatus.Status;
	}
	
	if (!NT_SUCCESS(Status)) 
	{
		KdPrint(("Ext2ReadSectors IO failed!!! Error code: %d(%x)\n", Status, Status));
		return Status;
	}
	
	return Status;
}

NTSTATUS
Ext2WriteSectors(IN PDEVICE_OBJECT pDeviceObject,
		 IN ULONG	DiskSector,
		 IN ULONG       SectorCount,
		 IN PVOID	Buffer)
{
	LARGE_INTEGER   sectorNumber;
	PIRP            irp;
	IO_STATUS_BLOCK ioStatus;
	KEVENT          event;
	NTSTATUS        Status;
	ULONG           sectorSize;
	
	KdPrint(("Ext2WriteSectors(pDeviceObject %xh, lba %xh, sectors: %xh Buffer %xh)\n",
		pDeviceObject,DiskSector, SectorCount, Buffer));
	
	sectorNumber.QuadPart = (LONGLONG)(DiskSector) * SECTOR_SIZE;

	KeInitializeEvent(&event, NotificationEvent, FALSE);
	
	sectorSize = SECTOR_SIZE * SectorCount;
	
	irp = IoBuildSynchronousFsdRequest(IRP_MJ_WRITE,
		pDeviceObject,
		Buffer,
		sectorSize,
		&sectorNumber,
		&event,
		&ioStatus );
	
	if (irp == NULL)
	{
		KdPrint(("Ext2WriteSectors: WRITE failed!!!\n"));
		return STATUS_UNSUCCESSFUL;
	}
	
	Status = IoCallDriver(pDeviceObject,
		irp);
	
	if (Status == STATUS_PENDING)
	{
		KeWaitForSingleObject(&event,
			Suspended,
			KernelMode,
			FALSE,
			NULL);
		Status = ioStatus.Status;
	}
	
	if (!NT_SUCCESS(Status))
	{
		KdPrint(("Ext2WriteSectors: IO failed!!! Error code: %d(%x)\n", Status, Status));
		return Status;
	}
	
	KdPrint(("Ext2WriteSectors: Block request succeeded.\n"));
	return Status;
}

NTSTATUS
Ext2ReadDisk(IN PDEVICE_OBJECT pDeviceObject,
	     IN ULONG lba,
	     IN ULONG offset,
	     IN ULONG Size,
	     IN PVOID Buffer)
{
	ULONG		Sectors;
	NTSTATUS	Status;
	PUCHAR		Buf;
	ULONG		off = 0;

	if (offset >= SECTOR_SIZE)
		lba += offset / SECTOR_SIZE;
	off = offset % SECTOR_SIZE;
	Sectors = (off + Size + SECTOR_SIZE - 1) / SECTOR_SIZE;

//      KdPrint(("Ext2ReadDisk: lba %xh, offset: %xh size: %xh Sectors:%xh.\n",
//              lba, offset, Size, Sectors));

	Buf = ExAllocatePool(NonPagedPool, Sectors * SECTOR_SIZE);
	if (!Buf)
	{
		KdPrint(("Ext2ReadDisk: no enough memory.\n"));
		return STATUS_UNSUCCESSFUL;
	}

	Status = Ext2ReadSectors(pDeviceObject, lba, Sectors, Buf);

	if (!NT_SUCCESS(Status))
	{
		KdPrint(("Ext2ReadDisk: Read Block Device error.\n"));
                ExFreePool(Buf);
		return Status;
	}

	RtlCopyMemory(Buffer, &Buf[off], Size);

	ExFreePool(Buf);

	return Status;
}


NTSTATUS
Ext2WriteDisk(IN PDEVICE_OBJECT pDeviceObject,
	      IN ULONG	lba,
 	      IN ULONG  offset,
	      IN ULONG  Size,
	      IN PVOID	Buffer)
{
	ULONG		Sectors;
	NTSTATUS	Status;
	PUCHAR		Buf;
	ULONG		off = 0;

	if (offset >= SECTOR_SIZE)
		lba += offset / SECTOR_SIZE;
	off = offset % SECTOR_SIZE;
	Sectors = (off + Size + SECTOR_SIZE - 1) / SECTOR_SIZE;

	Sectors = (off + Size + SECTOR_SIZE - 1) / SECTOR_SIZE;

	Buf = ExAllocatePool(NonPagedPool, Sectors * SECTOR_SIZE);
	if (!Buf)
	{
		KdPrint(("Ext2WriteDisk: no enough memory.\n"));
		return STATUS_UNSUCCESSFUL;
	}

	if (off)
	{
		Ext2ReadSectors(pDeviceObject, lba, 1, Buf);
	}

	if (((off + Size) % SECTOR_SIZE) != 0)
	{
		if ((Sectors != 1) || offset == 0)
			Ext2ReadSectors(pDeviceObject, lba + Sectors - 1, 1, Buf);
	}

	RtlCopyMemory(&Buf[off], Buffer, Size);

	Status = Ext2WriteSectors(pDeviceObject, lba, Sectors, Buf);

	if (!NT_SUCCESS(Status))
	{
		KdPrint(("Ext2WriteDisk: Write Block Device error.\n"));
	}

	ExFreePool(Buf);

	return Status;
}

NTSTATUS 
Ext2DiskIoControl (IN PDEVICE_OBJECT   pDeviceObject,
		   IN ULONG            IoctlCode,
		   IN PVOID            InputBuffer,
		   IN ULONG            InputBufferSize,
		   IN OUT PVOID        OutputBuffer,
		   IN OUT PULONG       OutputBufferSize)
{
	ULONG           OutBufferSize = 0;
	KEVENT          Event;
	PIRP            Irp;
	IO_STATUS_BLOCK IoStatus;
	NTSTATUS        Status;
	
	PAGED_CODE();
	
	ASSERT(pDeviceObject != NULL);
	
	if (OutputBufferSize)
	{
		OutBufferSize = *OutputBufferSize;
	}
	
	KeInitializeEvent(&Event, NotificationEvent, FALSE);
	
	Irp = IoBuildDeviceIoControlRequest(
		IoctlCode,
		pDeviceObject,
		InputBuffer,
		InputBufferSize,
		OutputBuffer,
		OutBufferSize,
		FALSE,
		&Event,
		&IoStatus
		);
	
	if (Irp == NULL)
	{
		KdPrint(("Ext2DiskIoControl: Building IRQ error!\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	
	Status = IoCallDriver(pDeviceObject, Irp);
	
	if (Status == STATUS_PENDING)
	{
		KeWaitForSingleObject(&Event, Executive, KernelMode, FALSE, NULL);
		Status = IoStatus.Status;
	}
	
	if (OutputBufferSize)
	{
		*OutputBufferSize = IoStatus.Information;
	}
	
	return Status;
}


NTSTATUS
Ext2ReadDiskOverrideVerify (IN PDEVICE_OBJECT pDeviceObject,
			    IN ULONG DiskSector, 
			    IN ULONG SectorCount,
			    IN OUT PUCHAR Buffer)
{
	PIO_STACK_LOCATION  IoStackLocation;
	LARGE_INTEGER sectorNumber;
	PIRP Irp;
	IO_STATUS_BLOCK IoStatus;
	KEVENT event;
	NTSTATUS Status;
	ULONG sectorSize;

	PAGED_CODE();

	sectorNumber.u.LowPart = DiskSector << 9;
	sectorNumber.u.HighPart = DiskSector >> 23;
	
	KeInitializeEvent (&event, NotificationEvent, FALSE);
	sectorSize = SECTOR_SIZE * SectorCount;
	
	Irp = IoBuildSynchronousFsdRequest (IRP_MJ_READ,
		pDeviceObject,
		Buffer,
		sectorSize,
		&sectorNumber, 
		&event, 
		&IoStatus);
	
	if (!Irp)
	{
		KdPrint(("Ext2ReadDiskOverrideVerify: IoBuildSynchronousFsdRequest failed\n"));
		return(STATUS_UNSUCCESSFUL);;
	}

	IoStackLocation = IoGetNextIrpStackLocation(Irp);
	IoStackLocation->Flags |= SL_OVERRIDE_VERIFY_VOLUME;
	
	Status = IoCallDriver (pDeviceObject, Irp);
	
	if (Status == STATUS_PENDING)
	{
		KeWaitForSingleObject (&event, Suspended, KernelMode, FALSE, NULL);
		Status = IoStatus.Status;
	}
	
	if (!NT_SUCCESS (Status))
	{
		KdPrint(("Ext2ReadDiskOverrideVerify: IO failed! Error code: %x\n", Status));
		return (Status);
	}

	return (STATUS_SUCCESS);
}


struct ext2_super_block *
Ext2LoadSuper(IN PDEVICE_OBJECT pDeviceObject)
{
	PVOID		Buffer;
	NTSTATUS	Status;
	
	Buffer = ExAllocatePool(NonPagedPool, 2 * SECTOR_SIZE);
	if (!Buffer)
	{
		KdPrint(("Ext2LoadSuper: no enough memory.\n"));
		return NULL;
	}

        Status = Ext2ReadDisk(pDeviceObject, 2, 0, SECTOR_SIZE * 2, Buffer);
	if (!NT_SUCCESS(Status))
	{
		KdPrint(("Ext2ReadDisk: Read Block Device error.\n"));
                ExFreePool(Buffer);
		return NULL;
	}

	return (struct ext2_super_block *)Buffer;
}

NTSTATUS
Ext2SaveSuper(IN PDEVICE_OBJECT pDeviceObject,
	      IN struct ext2_super_block * sb)
{
        return Ext2WriteDisk(pDeviceObject, 2, 0, SECTOR_SIZE * 2, (PVOID)sb);
}

struct ext2_group_desc *
Ext2LoadGroup(IN PEXT2_VCB vcb)
{
	ULONG		size;
	ULONG		lba;
	PVOID		Buffer;
	NTSTATUS	Status;
	PDEVICE_OBJECT pDeviceObject = vcb->TargetDeviceObject;
	struct ext2_super_block * sb = vcb->ext2_super_block;

	vcb->ext2_block = EXT2_MIN_BLOCK << sb->s_log_block_size;
	vcb->ext2_frag = EXT2_MIN_FRAG << sb->s_log_frag_size;

	vcb->ext2_groups = (sb->s_blocks_count - sb->s_first_data_block +
		sb->s_blocks_per_group - 1) / sb->s_blocks_per_group;

	size = sizeof(struct ext2_group_desc) * vcb->ext2_groups;

	Buffer = ExAllocatePool(NonPagedPool, size);
	if (!Buffer)
	{
		KdPrint(("Ext2LoadSuper: no enough memory.\n"));
		return NULL;
	}

	if (vcb->ext2_block == EXT2_MIN_BLOCK)
	{
		lba = 2 * (vcb->ext2_block / SECTOR_SIZE);
	}

	if (vcb->ext2_block > EXT2_MIN_BLOCK)
	{
		lba = vcb->ext2_block / SECTOR_SIZE;
	}

	Status = Ext2ReadDisk(pDeviceObject, lba, 0, size, Buffer);
	if (!NT_SUCCESS(Status))
	{
		KdPrint(("Ext2LoadGroup: Read Block Device error.\n"));
		ExFreePool(Buffer);
		return NULL;
	}

	return (struct ext2_group_desc *)Buffer;
}

NTSTATUS
Ext2SaveGroup(IN PEXT2_VCB vcb,
	      IN struct ext2_group_desc * gd)
{
	ULONG		lba;
	ULONG		size;
	PDEVICE_OBJECT pDeviceObject = vcb->TargetDeviceObject;

	size = vcb->ext2_groups * sizeof(struct ext2_group_desc);

	if (vcb->ext2_block == EXT2_MIN_BLOCK)
	{
		lba = 2 * (vcb->ext2_block / SECTOR_SIZE);
	}

	if (vcb->ext2_block > EXT2_MIN_BLOCK)
	{
		lba = vcb->ext2_block / SECTOR_SIZE;
	}

	return 	Ext2WriteDisk(pDeviceObject, lba, 0, size, (PVOID)gd);

}

BOOLEAN
Ext2GetInodeLba (IN PEXT2_VCB	vcb,
		 IN ULONG	inode,
		 OUT PULONG	lba,
		 OUT PULONG	offset)
{
	LONGLONG loc;

	if (inode < 1 || inode > vcb->ext2_super_block->s_inodes_count)
	{
		KdPrint(("Ext2GetInodeLba: Inode value %xh is invalid.\n",inode));
		*lba = 0;
		*offset = 0;
		return FALSE;
	}

	loc = (vcb->ext2_group_desc[(inode - 1) / vcb->ext2_super_block->s_inodes_per_group].bg_inode_table);
	loc = loc * vcb->ext2_block;
	loc = loc + ((inode - 1) % (vcb->ext2_super_block->s_inodes_per_group)) * sizeof(struct ext2_inode);

	*lba = (ULONG) (loc / SECTOR_SIZE);
	*offset = (ULONG) (loc % SECTOR_SIZE);

//	KdPrint(("Ext2GetInodeLba: inode=%xh lba=%xh offset=%xh\n",
//		inode, *lba, *offset));
	return TRUE;
}

BOOLEAN
Ext2LoadInode (IN PEXT2_VCB	vcb,
	       IN ULONG		inode,
	       IN struct ext2_inode *ext2_inode)
{
	ULONG		lba;
	ULONG		offset;
	NTSTATUS	Status;

	if (!Ext2GetInodeLba(vcb, inode, &lba, &offset))
	{
		KdPrint(("Ext2LoadInode: error get inode(%xh)'s addr.\n", inode));
		return FALSE;
	}

	Status = Ext2ReadDisk(vcb->TargetDeviceObject,
		lba,
		offset,
		sizeof(EXT2_INODE),
		(PVOID)ext2_inode);

	if (!NT_SUCCESS(Status))
	{
		return FALSE;
	}

	return TRUE;
}

BOOLEAN
Ext2SaveInode (IN PEXT2_VCB vcb,
	       IN ULONG inode,
	       IN struct ext2_inode *ext2_inode)
{
	ULONG		lba;
	ULONG		offset;
	NTSTATUS	Status;

	if (!Ext2GetInodeLba(vcb, inode, &lba, &offset))
	{
		KdPrint(("Ext2LoadInode: error get inode(%xh)'s addr.\n", inode));
		return FALSE;
	}

	Status = Ext2WriteDisk(vcb->TargetDeviceObject,
		lba,
		offset,
		sizeof(EXT2_INODE),
		(PVOID)ext2_inode);

	if (!NT_SUCCESS(Status))
	{
		return FALSE;
	}

	return TRUE;
}

ULONG Ext2GetBlock(IN PEXT2_VCB vcb,
		   ULONG dwContent,
		   ULONG Index,
		   int	 layer )
{
	ULONG		*pData = NULL;
	ULONG		i = 0, j = 0, temp = 1;
	ULONG		dwBlk = 0;
	LONGLONG	lba = 0;

	lba = (LONGLONG) dwContent;
	lba = lba * vcb->ext2_block / SECTOR_SIZE;

	if (layer == 0)
	{
		dwBlk = dwContent;
	}
	else if (layer <= 3)
	{
		pData = (ULONG *)ExAllocatePool(NonPagedPool,
					vcb->ext2_block);
		if (!pData)
		{
			KdPrint(("Ext2GetBlock: no enough memory.\n"));
			return dwBlk;
		}

		Ext2ReadDisk(vcb->TargetDeviceObject, (ULONG)lba, 0, vcb->ext2_block, (PVOID)pData);

		for (i=0; i<(ULONG)(layer - 1); i++)
		{
			temp *= (vcb->ext2_block/4);
		}

		i = Index / temp;
		j = Index % temp;

		dwBlk = Ext2GetBlock(vcb, pData[i], j, layer - 1);

		ExFreePool(pData);
	}
	
	return dwBlk;
}


ULONG Ext2BlockMap(IN PEXT2_VCB vcb,
		   IN struct ext2_inode* ext2_inode,
		   IN ULONG Index)
{
	ULONG dwSizes[3] = {vcb->ext2_block / 4, 0, 0};
	int   i;

	if (Index >= ext2_inode->i_blocks)
	{
		KdPrint(("Ext2BlockMap: error input paramters.\n"));
		return 0;
	}

	if (Index < 12)
	{
		return ext2_inode->i_block[Index];
	}

	Index -= 12;

	for (i = 0; i < 3; i++)
	{
		if (i > 0)
		{
			dwSizes[i] = vcb->ext2_block/4 * dwSizes[i-1];
		}

		if (Index < dwSizes[i])
		{
			return Ext2GetBlock(vcb, ext2_inode->i_block[i + 12], Index , i + 1); 
		}

		Index -= dwSizes[i];
	}

	return 0;
}


ULONG
Ext2ReadInode (IN PEXT2_VCB vcb,
	       IN struct ext2_inode* ext2_inode,
	       IN ULONG offset,
	       IN PVOID Buffer,
	       IN ULONG size )
{
	ULONG	dwBytes = 0;
	ULONG	nBeg, nEnd;
	ULONG	dwBlk;
	ULONG	off, dwSize;
	ULONG   i;
	LONGLONG lba;

	nBeg = offset / vcb->ext2_block;
	nEnd = (size + offset + vcb->ext2_block - 1) / vcb->ext2_block;

	for (i = nBeg; i < nEnd; i++)
	{
		dwBlk = Ext2BlockMap(vcb, ext2_inode, i);
		lba = (LONGLONG) dwBlk;
		lba = lba * vcb->ext2_block / SECTOR_SIZE;
		off = 0;
		dwSize = vcb->ext2_block;

		if (i == nBeg)
		{
			lba += (offset % vcb->ext2_block) / SECTOR_SIZE; 
			off = offset % SECTOR_SIZE;
			dwSize = vcb->ext2_block - (offset % vcb->ext2_block);
		}
		else if (i == nEnd)
		{
			dwSize = size - dwBytes;
		}
		Ext2ReadDisk(vcb->TargetDeviceObject,
			(ULONG)lba, off, dwSize, &((PUCHAR)Buffer)[dwBytes]);
		dwBytes += dwSize;

	}

	return dwBytes;
}

ULONG
Ext2WriteInode (IN PEXT2_VCB vcb,
	       IN struct ext2_inode* ext2_inode,
	       IN ULONG offset,
	       IN PVOID Buffer,
	       IN ULONG size )
{

	ULONG	dwBytes = 0;
	ULONG	nBeg, nEnd;
	ULONG	dwBlk;
	ULONG	off, dwSize;
	ULONG   i;
	LONGLONG lba;

	nBeg = offset / vcb->ext2_block;
	nEnd = (size + offset + vcb->ext2_block - 1) / vcb->ext2_block;

	for (i = nBeg; i < nEnd; i++)
	{
		dwBlk = Ext2BlockMap(vcb, ext2_inode, i);
		lba = (LONGLONG) dwBlk;
		lba = lba * vcb->ext2_block / SECTOR_SIZE;
		off = 0;
		dwSize = vcb->ext2_block;

		if (i == nBeg)
		{
			off = offset;
			dwSize = vcb->ext2_block - off;
		}
		else if (i == nEnd)
		{
			dwSize = size - dwBytes;
		}
		Ext2WriteDisk(vcb->TargetDeviceObject,
			(ULONG)lba, off, dwSize, &((PUCHAR)Buffer)[dwBytes]);
		dwBytes += dwSize;

	}

	return dwBytes;
}


PEXT2_IRP_CONTEXT
Ext2AllocateIrpContext (IN PDEVICE_OBJECT   DeviceObject,
			IN PIRP             Irp )
{
	PIO_STACK_LOCATION  IoStackLocation;
	PEXT2_IRP_CONTEXT    IrpContext;
	
	ASSERT(DeviceObject != NULL);
	ASSERT(Irp != NULL);
	
	IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
	
	IrpContext = (PEXT2_IRP_CONTEXT)ExAllocatePool(NonPagedPool, sizeof(EXT2_IRP_CONTEXT));
	
	if (!IrpContext)
	{
		return NULL;
	}
	
	IrpContext->Identifier.Type = ICX;
	IrpContext->Identifier.Size = sizeof(EXT2_IRP_CONTEXT);
	
	IrpContext->Irp = Irp;
	
	IrpContext->MajorFunction = IoStackLocation->MajorFunction;
	IrpContext->MinorFunction = IoStackLocation->MinorFunction;
	
	IrpContext->DeviceObject = DeviceObject;
	
	IrpContext->FileObject = IoStackLocation->FileObject;
	
	if (IrpContext->MajorFunction == IRP_MJ_FILE_SYSTEM_CONTROL ||
		IrpContext->MajorFunction == IRP_MJ_DEVICE_CONTROL ||
		IrpContext->MajorFunction == IRP_MJ_SHUTDOWN)
	{
		IrpContext->IsSynchronous = TRUE;
	}
	else if (IrpContext->MajorFunction == IRP_MJ_CLEANUP ||
		IrpContext->MajorFunction == IRP_MJ_CLOSE)
	{
		IrpContext->IsSynchronous = FALSE;
	}
	else
	{
		IrpContext->IsSynchronous = IoIsOperationSynchronous(Irp);
	}
	
	//
	// Temporary workaround for a bug in close that makes it reference a
	// fileobject when it is no longer valid.
	//
	if (IrpContext->MajorFunction == IRP_MJ_CLOSE)
	{
		IrpContext->IsSynchronous = TRUE;
	}
	
	IrpContext->IsTopLevel = (IoGetTopLevelIrp() == Irp);
	
	IrpContext->ExceptionInProgress = FALSE;
	
	return IrpContext;
}

VOID
Ext2FreeIrpContext (IN PEXT2_IRP_CONTEXT IrpContext)
{
	ASSERT(IrpContext != NULL);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	ExFreePool(IrpContext);
}

PEXT2_FCB
Ext2AllocateFcb (IN PEXT2_VCB            Vcb,
		 IN PUNICODE_STRING      FileName,
		 IN ULONG                IndexNumber,
		 IN ULONG		 inode,
		 IN ULONG		 dir_inode,
		 IN struct ext2_inode*   ext2_inode )
{
	PEXT2_FCB Fcb;
	
	Fcb = (PEXT2_FCB)ExAllocatePool(NonPagedPool, sizeof(EXT2_FCB));
	
	if (!Fcb)
	{
		return NULL;
	}
	
	Fcb->Identifier.Type = FCB;
	Fcb->Identifier.Size = sizeof(EXT2_FCB);
	
#ifndef EXT2_RO
	
	RtlZeroMemory(&Fcb->ShareAccess, sizeof(SHARE_ACCESS));
	
#endif
	
	FsRtlInitializeFileLock (
		&Fcb->FileLockAnchor,
		NULL,
		NULL );
	
	Fcb->OpenHandleCount = 0;
	Fcb->ReferenceCount = 0;
	
#ifndef EXT2_RO
	
	Fcb->IsPageFile = FALSE;
	
	Fcb->DeletePending = FALSE;
	
#endif
	
	Fcb->FileName.Length = 0;
	
	Fcb->FileName.MaximumLength = FileName->MaximumLength;
	
	Fcb->FileName.Buffer = (PWSTR) ExAllocatePool(
		NonPagedPool,
		Fcb->FileName.MaximumLength
		);
	
	if (!Fcb->FileName.Buffer)
	{
		ExFreePool(Fcb);
		return NULL;
	}
	
	RtlCopyUnicodeString(
		&Fcb->FileName,
		FileName );
	
#if DBG
	
	Fcb->DbgFileName = (PUCHAR) ExAllocatePool(
		NonPagedPool,
		Fcb->FileName.MaximumLength / sizeof(WCHAR) + 1
		);
	
	if (!Fcb->DbgFileName)
	{
		ExFreePool(Fcb->FileName.Buffer);
		ExFreePool(Fcb);
		return NULL;
	}
	
	Ext2WcharToChar (
		Fcb->DbgFileName,
		Fcb->FileName.Buffer,
		Fcb->FileName.Length / sizeof(WCHAR)
		);
	
	Fcb->DbgFileName[Fcb->FileName.Length / sizeof(WCHAR)] = 0;

#endif // DBG
	
	Fcb->FileAttributes = FILE_ATTRIBUTE_NORMAL;
	
	if (S_ISDIR(ext2_inode->i_mode))
	{
		Fcb->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
	}

#ifdef EXT2_RO

    {
        Fcb->FileAttributes |=  FILE_ATTRIBUTE_READONLY;
    }

#endif
	

	Fcb->IndexNumber.QuadPart = IndexNumber;
	
	Fcb->ext2_inode = ext2_inode;
	Fcb->inode = inode;
	Fcb->dir_inode = dir_inode;
	
	RtlZeroMemory(&Fcb->CommonFCBHeader, sizeof(FSRTL_COMMON_FCB_HEADER));
	
	//Fcb->CommonFCBHeader.NodeTypeCode = 
	Fcb->CommonFCBHeader.NodeByteSize = sizeof(FSRTL_COMMON_FCB_HEADER);
	Fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsNotPossible;
	Fcb->CommonFCBHeader.Resource = &(Fcb->MainResource);
	Fcb->CommonFCBHeader.PagingIoResource = &(Fcb->PagingIoResource);
	Fcb->CommonFCBHeader.AllocationSize.QuadPart = ext2_inode->i_size;
	Fcb->CommonFCBHeader.FileSize.QuadPart = ext2_inode->i_size;
	Fcb->CommonFCBHeader.ValidDataLength.QuadPart = ext2_inode->i_size;
	
	Fcb->SectionObject.DataSectionObject = NULL;
	Fcb->SectionObject.SharedCacheMap = NULL;
	Fcb->SectionObject.ImageSectionObject = NULL;
	
	ExInitializeResourceLite(&(Fcb->MainResource));
	ExInitializeResourceLite(&(Fcb->PagingIoResource));
	
	InsertTailList(&Vcb->FcbList, &Fcb->Next);
	
	return Fcb;
}

VOID
Ext2FreeFcb (IN PEXT2_FCB Fcb)
{
	ASSERT(Fcb != NULL);
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));

	FsRtlUninitializeFileLock(&Fcb->FileLockAnchor);

	ExDeleteResourceLite(&Fcb->MainResource);
	
	ExDeleteResourceLite(&Fcb->PagingIoResource);
	
	RemoveEntryList(&Fcb->Next);
	
	ExFreePool(Fcb->FileName.Buffer);
	
#if DBG
	ExFreePool(Fcb->DbgFileName);
#endif
	
	ExFreePool(Fcb->ext2_inode);
	
	ExFreePool(Fcb);
}

PEXT2_CCB
Ext2AllocateCcb (VOID)
{
	PEXT2_CCB Ccb;
	
	Ccb = (PEXT2_CCB)ExAllocatePool(NonPagedPool, sizeof(EXT2_CCB));
	
	if (!Ccb)
	{
		return NULL;
	}
	
	Ccb->Identifier.Type = CCB;
	Ccb->Identifier.Size = sizeof(EXT2_CCB);
	
	Ccb->CurrentByteOffset = 0;
	
	Ccb->DirectorySearchPattern.Length = 0;
	Ccb->DirectorySearchPattern.MaximumLength = 0;
	Ccb->DirectorySearchPattern.Buffer = 0;
	
	return Ccb;
}

VOID
Ext2FreeCcb (IN PEXT2_CCB Ccb)
{
	ASSERT(Ccb != NULL);
	
	ASSERT((Ccb->Identifier.Type == CCB) &&
		(Ccb->Identifier.Size == sizeof(EXT2_CCB)));
	
	if (Ccb->DirectorySearchPattern.Buffer != NULL)
	{
		ExFreePool(Ccb->DirectorySearchPattern.Buffer);
	}
	
	ExFreePool(Ccb);
}

VOID
Ext2FreeVcb (IN PEXT2_VCB Vcb )
{
	ASSERT(Vcb != NULL);
	
	ASSERT((Vcb->Identifier.Type == VCB) &&
		(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
	
	Ext2ClearVpbFlag(Vcb->Vpb, VPB_MOUNTED);
	
	ExAcquireResourceExclusiveLite(
		&Ext2Global->Resource,
		TRUE );
	
	RemoveEntryList(&Vcb->Next);
	
	ExReleaseResourceForThreadLite(
		&Ext2Global->Resource,
		ExGetCurrentResourceThread() );
	
	ExDeleteResourceLite(&Vcb->MainResource);
	
	ExDeleteResourceLite(&Vcb->PagingIoResource);
	
	if (Vcb->ext2_super_block)
		ExFreePool(Vcb->ext2_super_block);
	if (Vcb->ext2_group_desc)
		ExFreePool(Vcb->ext2_group_desc);
	
	IoDeleteDevice(Vcb->DeviceObject);
}

BOOLEAN
Ext2FastIoCheckIfPossible (
			  IN PFILE_OBJECT         FileObject,
			  IN PLARGE_INTEGER       FileOffset,
			  IN ULONG                Length,
			  IN BOOLEAN              Wait,
			  IN ULONG                LockKey,
			  IN BOOLEAN              CheckForReadOperation,
			  OUT PIO_STATUS_BLOCK    IoStatus,
			  IN PDEVICE_OBJECT       DeviceObject
			  )
{
	BOOLEAN          Status = FALSE;
	PEXT2_FCB        Fcb;
	LARGE_INTEGER    lLength;
	
	lLength.QuadPart = Length;
	
	PAGED_CODE();
	
	__try
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				Status = FALSE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				Status = FALSE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
			
			if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				Status = FALSE;
				__leave;
			}
			
			FsRtlEnterFileSystem();
			
			if (CheckForReadOperation)
			{
				Status = FsRtlFastCheckLockForRead(
					&Fcb->FileLockAnchor,
					FileOffset,
					&lLength,
					LockKey,
					FileObject,
					PsGetCurrentProcess());
			}
			else
			{
				Status = FsRtlFastCheckLockForWrite(
					&Fcb->FileLockAnchor,
					FileOffset,
					&lLength,
					LockKey,
					FileObject,
					PsGetCurrentProcess());
			}
#if DBG			
			KdPrint(("Ext2FastIOCheckPossible: %-16.16s %-31s %s\n",
				PsGetCurrentProcess()->ImageFileName,
				"FASTIO_CHECK_IF_POSSIBLE",
				Fcb->DbgFileName));
#endif			
			KdPrint(("Ext2FastIIOCheckPossible: Offset: %I64u Length: %u Key: %u %s %s\n",
				FileOffset->QuadPart,
				Length,
				LockKey,
				(CheckForReadOperation ? "CheckForReadOperation:" : "CheckForWriteOperation:"),
				(Status ? "Succeeded" : "Failed")));
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			Status = FALSE;
		}
	}
	__finally
	{
		FsRtlExitFileSystem();
	}
	
	return Status;
}


#if DBG
BOOLEAN
Ext2FastIoRead (IN PFILE_OBJECT         FileObject,
	       IN PLARGE_INTEGER       FileOffset,
	       IN ULONG                Length,
	       IN BOOLEAN              Wait,
	       IN ULONG                LockKey,
	       OUT PVOID               Buffer,
	       OUT PIO_STATUS_BLOCK    IoStatus,
	       IN PDEVICE_OBJECT       DeviceObject)
{
	BOOLEAN     Status;
	PEXT2_FCB    Fcb;
	
	PAGED_CODE();
	
	Fcb = (PEXT2_FCB) FileObject->FsContext;
	
	ASSERT(Fcb != NULL);
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
	
	Status = FsRtlCopyRead (
		FileObject, FileOffset,	Length,	Wait,
		LockKey, Buffer, IoStatus, DeviceObject);
	
	return Status;
}

#endif /* DBG */


BOOLEAN
Ext2FastIoQueryBasicInfo (IN PFILE_OBJECT             FileObject,
			  IN BOOLEAN                  Wait,
			  OUT PFILE_BASIC_INFORMATION Buffer,
			  OUT PIO_STATUS_BLOCK        IoStatus,
			  IN PDEVICE_OBJECT           DeviceObject)
{
	BOOLEAN     Status = FALSE;
	PEXT2_FCB Fcb;
	BOOLEAN     FcbMainResourceAcquired = FALSE;
	
	PAGED_CODE();
	__try
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
				Status = TRUE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
#if DBG			
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				PsGetCurrentProcess()->ImageFileName,
				"FASTIO_QUERY_BASIC_INFO",
				Fcb->DbgFileName
				));
#endif			
			FsRtlEnterFileSystem();
			
			if (!ExAcquireResourceSharedLite(
				&Fcb->MainResource,
				Wait))
			{
				Status = FALSE;
				__leave;
			}
			
			FcbMainResourceAcquired = TRUE;
			
			RtlZeroMemory(Buffer, sizeof(FILE_BASIC_INFORMATION));
			
			/*
			typedef struct _FILE_BASIC_INFORMATION {
			LARGE_INTEGER   CreationTime;
			LARGE_INTEGER   LastAccessTime;
			LARGE_INTEGER   LastWriteTime;
			LARGE_INTEGER   ChangeTime;
			ULONG           FileAttributes;
			} FILE_BASIC_INFORMATION, *PFILE_BASIC_INFORMATION;
			*/

			Buffer->CreationTime = Ext2SysTime(Fcb->ext2_inode->i_ctime + TIMEZONE);
			Buffer->LastAccessTime = Ext2SysTime(Fcb->ext2_inode->i_atime + TIMEZONE);
			Buffer->LastWriteTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
			Buffer->ChangeTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
			
			
			Buffer->FileAttributes = Fcb->FileAttributes;
			
			IoStatus->Information = sizeof(FILE_BASIC_INFORMATION);
			
			IoStatus->Status = STATUS_SUCCESS;
			
			Status =  TRUE;
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			IoStatus->Status = GetExceptionCode();
			Status = TRUE;
		}
	}

	__finally
	{
		if (FcbMainResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread()
				);
		}
		
		FsRtlExitFileSystem();
	}
	
	
	
	if (Status == FALSE)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: FALSE ***\n",
			PsGetCurrentProcess()->ImageFileName,
			"FASTIO_QUERY_BASIC_INFO"
			));
	}
	else if (IoStatus->Status != STATUS_SUCCESS)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
			PsGetCurrentProcess()->ImageFileName,
			"FASTIO_QUERY_BASIC_INFO",
			Ext2NtStatusToString(IoStatus->Status),
			IoStatus->Status
			));
	}
	
	return Status;

// Matt Wu 2001-09-29
/*	
	if (DeviceObject == Ext2Global->DeviceObject)
	{
		return FALSE;
	}
	
	Fcb = (PEXT2_FCB) FileObject->FsContext;
	
	ASSERT(Fcb != NULL);
	
	if (Fcb->Identifier.Type == VCB)
	{
		return FALSE;
	}
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));

	FsRtlEnterFileSystem();
	
	if (!ExAcquireResourceSharedLite(
		&Fcb->MainResource,Wait ))
	{
		FsRtlExitFileSystem();
		return FALSE;
	}
	
	RtlZeroMemory(Buffer, sizeof(FILE_BASIC_INFORMATION));
	
	Buffer->CreationTime = Ext2SysTime(Fcb->ext2_inode->i_ctime + TIMEZONE);
	Buffer->LastAccessTime = Ext2SysTime(Fcb->ext2_inode->i_atime + TIMEZONE);
	Buffer->LastWriteTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
	Buffer->ChangeTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
	Buffer->FileAttributes = Fcb->FileAttributes;
	
	IoStatus->Information = sizeof(FILE_BASIC_INFORMATION);
	
	IoStatus->Status = STATUS_SUCCESS;
	
	ExReleaseResourceForThreadLite(
		&Fcb->MainResource,
		ExGetCurrentResourceThread());
	
	FsRtlExitFileSystem();
	
	return TRUE;
*/

}

BOOLEAN
Ext2FastIoQueryStandardInfo (
			    IN PFILE_OBJECT                 FileObject,
			    IN BOOLEAN                      Wait,
			    OUT PFILE_STANDARD_INFORMATION  Buffer,
			    OUT PIO_STATUS_BLOCK            IoStatus,
			    IN PDEVICE_OBJECT               DeviceObject
			    )
{
	
	BOOLEAN     Status = FALSE;
	PEXT2_FCB   Fcb;
	BOOLEAN     FcbMainResourceAcquired = FALSE;
	
	PAGED_CODE();
	
	__try
		
		
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
				Status = TRUE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
#if DBG			
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				PsGetCurrentProcess()->ImageFileName,
				"FASTIO_QUERY_STANDARD_INFO",
				Fcb->DbgFileName ));
#endif			
			FsRtlEnterFileSystem();
			
			if (!ExAcquireResourceSharedLite(
				&Fcb->MainResource,
				Wait		))
			{
				Status = FALSE;
				__leave;
			}
			
			FcbMainResourceAcquired = TRUE;
			
			RtlZeroMemory(Buffer, sizeof(FILE_STANDARD_INFORMATION));
			
			/*
			typedef struct _FILE_STANDARD_INFORMATION {
			LARGE_INTEGER   AllocationSize;
			LARGE_INTEGER   EndOfFile;
			ULONG           NumberOfLinks;
			BOOLEAN         DeletePending;
			BOOLEAN         Directory;
			} FILE_STANDARD_INFORMATION, *PFILE_STANDARD_INFORMATION;
			*/
			
			Buffer->AllocationSize.QuadPart =
				Fcb->ext2_inode->i_size;
			Buffer->EndOfFile.QuadPart =
				Fcb->ext2_inode->i_size;			
			Buffer->NumberOfLinks = Fcb->ext2_inode->i_links_count;
			
#ifndef EXT2_RO
			Buffer->DeletePending = Fcb->DeletePending;
#else
			Buffer->DeletePending = FALSE;
#endif
			
			if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				Buffer->Directory = TRUE;
			}
			else
			{
				Buffer->Directory = FALSE;
			}
			
			IoStatus->Information = sizeof(FILE_STANDARD_INFORMATION);
			
			IoStatus->Status = STATUS_SUCCESS;
			
			Status =  TRUE;
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			IoStatus->Status = GetExceptionCode();
			Status = TRUE;
		}
	}
	__finally
	{
		if (FcbMainResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread()
				);
		}
		
		FsRtlExitFileSystem();
	}
	
	if (Status == FALSE)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: FALSE ***\n",
			PsGetCurrentProcess()->ImageFileName,
			"FASTIO_QUERY_STANDARD_INFO"
			));
	}
	else if (IoStatus->Status != STATUS_SUCCESS)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
			PsGetCurrentProcess()->ImageFileName,
			"FASTIO_QUERY_STANDARD_INFO",
			Ext2NtStatusToString(IoStatus->Status),
			IoStatus->Status
			));
	}
	
	return Status;
}

BOOLEAN
Ext2FastIoLock (
	       IN PFILE_OBJECT         FileObject,
	       IN PLARGE_INTEGER       FileOffset,
	       IN PLARGE_INTEGER       Length,
	       IN PEPROCESS            ProcessId,
	       IN ULONG                Key,
	       IN BOOLEAN              FailImmediately,
	       IN BOOLEAN              ExclusiveLock,
	       OUT PIO_STATUS_BLOCK    IoStatus,
	       IN PDEVICE_OBJECT       DeviceObject
	       )
{
	BOOLEAN     Status = FALSE;
	PEXT2_FCB   Fcb;
	
	PAGED_CODE();
	
	__try
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
				Status = TRUE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
			
			if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
#if DBG			
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				ProcessId->ImageFileName,
				"FASTIO_LOCK",
				Fcb->DbgFileName		));
#endif			
			KdPrint((
				DRIVER_NAME ": Offset: %I64u Length: %I64u Key: %u %s%s\n",
				FileOffset->QuadPart,
				Length->QuadPart,
				Key,
				(FailImmediately ? "FailImmediately " : ""),
				(ExclusiveLock ? "ExclusiveLock " : "")	));
			
			if (Fcb->CommonFCBHeader.IsFastIoPossible != FastIoIsQuestionable)
			{
#if DBG
				KdPrint((
					DRIVER_NAME ": %-16.16s %-31s %s\n",
					ProcessId->ImageFileName,
					"FastIoIsQuestionable",
					Fcb->DbgFileName		));
#endif				
				Fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsQuestionable;
			}
			
			FsRtlEnterFileSystem();
			
			Status = FsRtlPrivateLock(
				&Fcb->FileLockAnchor,
				FileObject,
				FileOffset,
				Length,
				ProcessId,
				Key,
				FailImmediately,
				ExclusiveLock,
				IoStatus,
				NULL,
				NULL,
				NULL);
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			IoStatus->Status = GetExceptionCode();
			Status = TRUE;
		}
	}

	__finally
	{
		FsRtlExitFileSystem();
	}
	
	if (Status == FALSE)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: FALSE ***\n",
			ProcessId->ImageFileName,
			"FASTIO_LOCK"
			));
	}
	else if (IoStatus->Status != STATUS_SUCCESS)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
			ProcessId->ImageFileName,
			"FASTIO_LOCK",
			Ext2NtStatusToString(IoStatus->Status),
			IoStatus->Status
			));
	}
	
	return Status;
}

BOOLEAN
Ext2FastIoUnlockSingle (
		       IN PFILE_OBJECT         FileObject,
		       IN PLARGE_INTEGER       FileOffset,
		       IN PLARGE_INTEGER       Length,
		       IN PEPROCESS            ProcessId,
		       IN ULONG                Key,
		       OUT PIO_STATUS_BLOCK    IoStatus,
		       IN PDEVICE_OBJECT       DeviceObject
		       )
{
	BOOLEAN     Status = FALSE;
	PEXT2_FCB   Fcb;
	
	PAGED_CODE();
	
	__try
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
				Status = TRUE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
			
			if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
#if DBG			
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				ProcessId->ImageFileName,
				"FASTIO_UNLOCK_SINGLE",
				Fcb->DbgFileName		));
#endif			
			KdPrint((
				DRIVER_NAME ": Offset: %I64u Length: %I64u Key: %u\n",
				FileOffset->QuadPart,
				Length->QuadPart,
				Key		));
			
			FsRtlEnterFileSystem();
			
			IoStatus->Status = FsRtlFastUnlockSingle(
				&Fcb->FileLockAnchor,
				FileObject,
				FileOffset,
				Length,
				ProcessId,
				Key,
				NULL,
				NULL);                      
			
			IoStatus->Information = 0;
			
			Status =  TRUE;
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			IoStatus->Status = GetExceptionCode();
			Status = TRUE;
		}
	}
	__finally
	{
		FsRtlExitFileSystem();
	}
	
	if (Status == FALSE)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: FALSE ***\n",
			ProcessId->ImageFileName,
			"FASTIO_UNLOCK_SINGLE"
			));
	}
	else if (IoStatus->Status != STATUS_SUCCESS)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
			ProcessId->ImageFileName,
			"FASTIO_UNLOCK_SINGLE",
			Ext2NtStatusToString(IoStatus->Status),
			IoStatus->Status
			));
	}
	
	return Status;
}

BOOLEAN
Ext2FastIoUnlockAll (
		    IN PFILE_OBJECT         FileObject,
		    IN PEPROCESS            ProcessId,
		    OUT PIO_STATUS_BLOCK    IoStatus,
		    IN PDEVICE_OBJECT       DeviceObject
		    )
{
	BOOLEAN     Status = FALSE;
	PEXT2_FCB   Fcb;
	
	PAGED_CODE();
	
	__try
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
				Status = TRUE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
			
			if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
#if DBG			
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				ProcessId->ImageFileName,
				"FASTIO_UNLOCK_ALL",
				Fcb->DbgFileName
				));
#endif			
			FsRtlEnterFileSystem();
			
			IoStatus->Status = FsRtlFastUnlockAll(
				&Fcb->FileLockAnchor,
				FileObject,
				ProcessId,
				NULL
				);
			
			IoStatus->Information = 0;
			
			Status =  TRUE;
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			IoStatus->Status = GetExceptionCode();
			Status = TRUE;
		}
	}
	__finally
	{
		FsRtlExitFileSystem();
	}
	
	if (Status == FALSE)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: FALSE ***\n",
			ProcessId->ImageFileName,
			"FASTIO_UNLOCK_ALL"
			));
	}
	else if (IoStatus->Status != STATUS_SUCCESS)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
			ProcessId->ImageFileName,
			"FASTIO_UNLOCK_ALL",
			Ext2NtStatusToString(IoStatus->Status),
			IoStatus->Status
			));
	}
	
	return Status;
}

BOOLEAN
Ext2FastIoUnlockAllByKey (
			 IN PFILE_OBJECT         FileObject,
			 IN PEPROCESS            ProcessId,
			 IN ULONG                Key,
			 OUT PIO_STATUS_BLOCK    IoStatus,
			 IN PDEVICE_OBJECT       DeviceObject
			 )
{
	BOOLEAN     Status = FALSE;
	PEXT2_FCB   Fcb;
	
	PAGED_CODE();
	
	__try
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
				Status = TRUE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
			
			if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
#if DBG			
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				ProcessId->ImageFileName,
				"FASTIO_UNLOCK_ALL_BY_KEY",
				Fcb->DbgFileName
				));
#endif			
			KdPrint((
				DRIVER_NAME ": Key: %u\n",
				Key
				));
			
			FsRtlEnterFileSystem();
			
			IoStatus->Status = FsRtlFastUnlockAllByKey(
				&Fcb->FileLockAnchor,
				FileObject,
				ProcessId,
				Key,
				NULL
				);  
			
			IoStatus->Information = 0;
			
			Status =  TRUE;
		}

		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			IoStatus->Status = GetExceptionCode();
			Status = TRUE;
		}
	}

	__finally
	{
		FsRtlExitFileSystem();
	}
	
	if (Status == FALSE)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: FALSE ***\n",
			ProcessId->ImageFileName,
			"FASTIO_UNLOCK_ALL_BY_KEY"
			));
	}
	else if (IoStatus->Status != STATUS_SUCCESS)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
			ProcessId->ImageFileName,
			"FASTIO_UNLOCK_ALL_BY_KEY",
			Ext2NtStatusToString(IoStatus->Status),
			IoStatus->Status
			));
	}
	
	return Status;
}


BOOLEAN
Ext2FastIoQueryNetworkOpenInfo (
	 IN PFILE_OBJECT                     FileObject,
	 IN BOOLEAN                          Wait,
	 OUT PFILE_NETWORK_OPEN_INFORMATION  Buffer,
	 OUT PIO_STATUS_BLOCK                IoStatus,
	 IN PDEVICE_OBJECT                   DeviceObject )
{
	BOOLEAN     Status = FALSE;
	PEXT2_FCB   Fcb;
	BOOLEAN     FcbMainResourceAcquired = FALSE;
	
	PAGED_CODE();
	
	__try
		
		
	{
		__try
		{
			if (DeviceObject == Ext2Global->DeviceObject)
			{
				IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
				Status = TRUE;
				__leave;
			}
			
			Fcb = (PEXT2_FCB) FileObject->FsContext;
			
			ASSERT(Fcb != NULL);
			
			if (Fcb->Identifier.Type == VCB)
			{
				IoStatus->Status = STATUS_INVALID_PARAMETER;
				Status = TRUE;
				__leave;
			}
			
			ASSERT((Fcb->Identifier.Type == FCB) &&
				(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
#if DBG			
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				PsGetCurrentProcess()->ImageFileName,
				"FASTIO_QUERY_NETWORK_OPEN_INFO",
				Fcb->DbgFileName
				));
#endif			
			FsRtlEnterFileSystem();
			
			if (!ExAcquireResourceSharedLite(
				&Fcb->MainResource,
				Wait
				))
			{
				Status = FALSE;
				__leave;
			}
			
			FcbMainResourceAcquired = TRUE;
			
			RtlZeroMemory(Buffer, sizeof(FILE_NETWORK_OPEN_INFORMATION));
			
			/*
			typedef struct _FILE_NETWORK_OPEN_INFORMATION {
			LARGE_INTEGER   CreationTime;
			LARGE_INTEGER   LastAccessTime;
			LARGE_INTEGER   LastWriteTime;
			LARGE_INTEGER   ChangeTime;
			LARGE_INTEGER   AllocationSize;
			LARGE_INTEGER   EndOfFile;
			ULONG           FileAttributes;
			} FILE_NETWORK_OPEN_INFORMATION, *PFILE_NETWORK_OPEN_INFORMATION;
			*/
			
			Buffer->CreationTime = Ext2SysTime(Fcb->ext2_inode->i_ctime + TIMEZONE);
			Buffer->LastAccessTime = Ext2SysTime(Fcb->ext2_inode->i_atime + TIMEZONE);
			Buffer->LastWriteTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
			Buffer->ChangeTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
			Buffer->FileAttributes = Fcb->FileAttributes;
			Buffer->AllocationSize.QuadPart =
				Fcb->ext2_inode->i_size;
			Buffer->EndOfFile.QuadPart =
				Fcb->ext2_inode->i_size;			
			
			Buffer->FileAttributes = Fcb->FileAttributes;
			
			IoStatus->Information = sizeof(FILE_NETWORK_OPEN_INFORMATION);
			
			IoStatus->Status = STATUS_SUCCESS;
			
			Status =  TRUE;
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			IoStatus->Status = GetExceptionCode();
			Status = TRUE;
		}
	}
	__finally
	{
		if (FcbMainResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread()
				);
		}
		
		FsRtlExitFileSystem();
	}
	
	
	
	if (Status == FALSE)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: FALSE ***\n",
			PsGetCurrentProcess()->ImageFileName,
			"FASTIO_QUERY_NETWORK_OPEN_INFO"
			));
	}
	else if (IoStatus->Status != STATUS_SUCCESS)
	{
		KdPrint((
			DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
			PsGetCurrentProcess()->ImageFileName,
			"FASTIO_QUERY_NETWORK_OPEN_INFO",
			Ext2NtStatusToString(IoStatus->Status),
			IoStatus->Status
			));
	}
	
	return Status;
}


NTSTATUS
Ext2LockControl (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT  DeviceObject;
	BOOLEAN         CompleteRequest;
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	PFILE_OBJECT    FileObject;
	PEXT2_FCB       Fcb;
	PIRP            Irp;
	
	PAGED_CODE();
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			CompleteRequest = TRUE;
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		FileObject = IrpContext->FileObject;
		
		Fcb = (PEXT2_FCB) FileObject->FsContext;
		
		ASSERT(Fcb != NULL);
		
		if (Fcb->Identifier.Type == VCB)
		{
			CompleteRequest = TRUE;
			Status = STATUS_INVALID_PARAMETER;
			__leave;
		}
		
		ASSERT((Fcb->Identifier.Type == FCB) &&
			(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
		
		if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			CompleteRequest = TRUE;
			Status = STATUS_INVALID_PARAMETER;
			__leave;
		}
		
		Irp = IrpContext->Irp;
		
		//
		// While the file has any byte range locks we set IsFastIoPossible to
		// FastIoIsQuestionable so that the FastIoCheckIfPossible function is
		// called to check the locks for any fast I/O read/write operation.
		//
		if (Fcb->CommonFCBHeader.IsFastIoPossible != FastIoIsQuestionable)
		{
#if DBG
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s %s\n",
				PsGetCurrentProcess()->ImageFileName,
				"FastIoIsQuestionable",
				Fcb->DbgFileName
				));
#endif			
			Fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsQuestionable;
		}
		
		//
		// FsRtlProcessFileLock acquires FileObject->FsContext->Resource while
		// modifying the file locks and calls IoCompleteRequest when it's done.
		//
		
		CompleteRequest = FALSE;
		
		Status = FsRtlProcessFileLock(
			&Fcb->FileLockAnchor,
			Irp,
			NULL		);
		
		if (Status != STATUS_SUCCESS)
		{
			KdPrint((
				DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
				PsGetCurrentProcess()->ImageFileName,
				"IRP_MJ_LOCK_CONTROL",
				Ext2NtStatusToString(Status),
				Status			));
		}
	}

	__finally
	{
		if (!IrpContext->ExceptionInProgress)
		{
			if (CompleteRequest)
			{
				IrpContext->Irp->IoStatus.Status = Status;
				
				Ext2CompleteRequest(
					IrpContext->Irp,
					(CCHAR)
					(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
					);
			}
			
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}


/*
BOOLEAN
Ext2FastIoQueryStandardInfo (IN PFILE_OBJECT                 FileObject,
			     IN BOOLEAN                      Wait,
			     OUT PFILE_STANDARD_INFORMATION  Buffer,
			     OUT PIO_STATUS_BLOCK            IoStatus,
			     IN PDEVICE_OBJECT               DeviceObject)
{
	PEXT2_FCB Fcb;
	
	PAGED_CODE();
	
	if (DeviceObject == Ext2Global->DeviceObject)
	{
		return FALSE;
	}
	
	Fcb = (PEXT2_FCB) FileObject->FsContext;
	
	ASSERT(Fcb != NULL);
	
	if (Fcb->Identifier.Type == VCB)
	{
		return FALSE;
	}
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));

	FsRtlEnterFileSystem();
	
	if (!ExAcquireResourceSharedLite(
		&Fcb->MainResource,
		Wait
		))
	{
		FsRtlExitFileSystem();
		return FALSE;
	}
	
	RtlZeroMemory(Buffer, sizeof(FILE_STANDARD_INFORMATION));
	Buffer->AllocationSize.QuadPart = Fcb->ext2_inode->i_size;
	Buffer->EndOfFile.QuadPart = Fcb->ext2_inode->i_size;
	Buffer->NumberOfLinks = Fcb->ext2_inode->i_links_count;
	
#ifndef EXT2_RO
	Buffer->DeletePending = Fcb->DeletePending;
#else
	Buffer->DeletePending = FALSE;
#endif
	
	if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	{
		Buffer->Directory = TRUE;
	}
	else
	{
		Buffer->Directory = FALSE;
	}
	
	IoStatus->Information = sizeof(FILE_STANDARD_INFORMATION);
	IoStatus->Status = STATUS_SUCCESS;
	ExReleaseResourceForThreadLite(
		&Fcb->MainResource,
		ExGetCurrentResourceThread());
	FsRtlExitFileSystem();

	return TRUE;
}

BOOLEAN
Ext2FastIoQueryNetworkOpenInfo (IN PFILE_OBJECT                     FileObject,
				IN BOOLEAN                          Wait,
				OUT PFILE_NETWORK_OPEN_INFORMATION  Buffer,
				OUT PIO_STATUS_BLOCK                IoStatus,
				IN PDEVICE_OBJECT                   DeviceObject)
{
	PEXT2_FCB Fcb;
	
	PAGED_CODE();
	
	if (DeviceObject == Ext2Global->DeviceObject)
	{
		return FALSE;
	}
	
	Fcb = (PEXT2_FCB) FileObject->FsContext;
	
	ASSERT(Fcb != NULL);
	
	if (Fcb->Identifier.Type == VCB)
	{
		return FALSE;
	}
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
	FsRtlEnterFileSystem();
	
	if (!ExAcquireResourceSharedLite(
		&Fcb->MainResource, Wait ))
	{
		FsRtlExitFileSystem();
		return FALSE;
	}
	
	RtlZeroMemory(Buffer, sizeof(FILE_NETWORK_OPEN_INFORMATION));
	Buffer->CreationTime = Ext2SysTime(Fcb->ext2_inode->i_ctime + TIMEZONE);
	Buffer->LastAccessTime = Ext2SysTime(Fcb->ext2_inode->i_atime + TIMEZONE);
	Buffer->LastWriteTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
	Buffer->ChangeTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
	Buffer->AllocationSize.QuadPart = Fcb->ext2_inode->i_size;
	Buffer->EndOfFile.QuadPart = Fcb->ext2_inode->i_size;
	Buffer->FileAttributes = Fcb->FileAttributes;
	IoStatus->Information = sizeof(FILE_NETWORK_OPEN_INFORMATION);
	IoStatus->Status = STATUS_SUCCESS;
	ExReleaseResourceForThreadLite(
		&Fcb->MainResource,
		ExGetCurrentResourceThread());
	
	FsRtlExitFileSystem();
	
	return TRUE;
}

*/


BOOLEAN
Ext2AcquireForLazyWrite (IN PVOID    Context,
			 IN BOOLEAN  Wait)
{
	//
	// On a readonly filesystem this function still has to exist but it
	// doesn't need to do anything.
	
	PEXT2_FCB    Fcb;
#ifndef EXT2_RO
	BOOLEAN     Status;
#endif
	
	PAGED_CODE();
	
	Fcb = (PEXT2_FCB) Context;
	
	ASSERT(Fcb != NULL);
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
	
#ifdef EXT2_RO
	return TRUE;
#else
	if (ExAcquireResourceExclusiveLite(
		&Fcb->MainResource,Wait)) {
		Status = TRUE;
	} else {
		Status = FALSE;
	}
	
	return Status;
#endif
}

VOID
Ext2ReleaseFromLazyWrite (IN PVOID Context)
{
	//
	// On a readonly filesystem this function still has to exist but it
	// doesn't need to do anything.
	PEXT2_FCB Fcb;
	
	PAGED_CODE();
	
	Fcb = (PEXT2_FCB) Context;
	
	ASSERT(Fcb != NULL);
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
	
#ifdef EXT2_RO
	return;
#else
	
	ExReleaseResourceForThreadLite(
		&Fcb->MainResource,
		ExGetCurrentResourceThread());
#endif
}

BOOLEAN
Ext2AcquireForReadAhead (IN PVOID    Context,
			 IN BOOLEAN  Wait)
{
	PEXT2_FCB    Fcb;
	BOOLEAN     Status;
	
	PAGED_CODE();
	
	Fcb = (PEXT2_FCB) Context;
	
	ASSERT(Fcb != NULL);
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
	
	if (ExAcquireResourceSharedLite(
		&Fcb->MainResource, Wait))
	{
		Status = TRUE;
	}
	else
	{
		Status = FALSE;
	}
	
	return Status;
}

VOID
Ext2ReleaseFromReadAhead (IN PVOID Context)
{
	PEXT2_FCB Fcb;
	
	PAGED_CODE();
	
	Fcb = (PEXT2_FCB) Context;
	
	ASSERT(Fcb != NULL);
	
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
	
	ExReleaseResourceForThreadLite(
		&Fcb->MainResource,
		ExGetCurrentResourceThread());
}


NTSTATUS
Ext2Close (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT  DeviceObject;
	NTSTATUS        Status = STATUS_SUCCESS;
	PEXT2_VCB       Vcb;
	BOOLEAN         VcbResourceAcquired = FALSE;
	PFILE_OBJECT    FileObject;
	PEXT2_FCB       Fcb;
	BOOLEAN         FcbResourceAcquired = FALSE;
	PEXT2_CCB       Ccb;
	BOOLEAN         FreeVcb = FALSE;

	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;

		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_SUCCESS;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		if (!ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			IrpContext->IsSynchronous ))
		{
			KdPrint(("Ext2Close: PENDING ... Vcb: %xh/%xh\n", Vcb->OpenFileHandleCount, Vcb->ReferenceCount));
			Status = STATUS_PENDING;
			__leave;
		}
		
		VcbResourceAcquired = TRUE;
		
		FileObject = IrpContext->FileObject;
		
		Fcb = (PEXT2_FCB) FileObject->FsContext;
		
		ASSERT(Fcb != NULL);
		
		if (Fcb->Identifier.Type == VCB)
		{
			if (Vcb->Flags & VCB_VOLUME_LOCKED)
			{
				Vcb->Flags &= ~VCB_VOLUME_LOCKED;
				
				Ext2ClearVpbFlag(Vcb->Vpb, VPB_LOCKED);
			}
			
			Vcb->ReferenceCount--;
			
			if (!Vcb->ReferenceCount && Vcb->Flags & VCB_DISMOUNT_PENDING)
			{
				FreeVcb = TRUE;
			}
			
			Status = STATUS_SUCCESS;
			
			__leave;
		}
		
		ASSERT((Fcb->Identifier.Type == FCB) &&
			(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
		
#ifndef EXT2_RO
		if (!Fcb->IsPageFile)
#endif
		{
			if (!ExAcquireResourceExclusiveLite(
				&Fcb->MainResource,
				IrpContext->IsSynchronous ))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			FcbResourceAcquired = TRUE;
		}
		
		Ccb = (PEXT2_CCB) FileObject->FsContext2;
		
		ASSERT(Ccb != NULL);
		
		ASSERT((Ccb->Identifier.Type == CCB) &&
			(Ccb->Identifier.Size == sizeof(EXT2_CCB)));
		
		Fcb->ReferenceCount--;
		
		Vcb->ReferenceCount--;

		if (!Vcb->ReferenceCount && Vcb->Flags & VCB_DISMOUNT_PENDING)
		{
			FreeVcb = TRUE;
		}
		

		Ext2FreeCcb(Ccb);
		
		if (!Fcb->ReferenceCount)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread() );
			
			FcbResourceAcquired = FALSE;
			
			Ext2FreeFcb(Fcb);
		}
		
		Status = STATUS_SUCCESS;
    }
    __finally
    {
	    if (VcbResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Vcb->MainResource,
			    ExGetCurrentResourceThread()  );
	    }
	    
	    if (FcbResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Fcb->MainResource,
			    ExGetCurrentResourceThread() );
	    }
	    
	    if (!IrpContext->ExceptionInProgress)
	    {
		    if (Status == STATUS_PENDING)
		    {
		    
			    Status = STATUS_SUCCESS;
			    
			    if (IrpContext->Irp != NULL)
			    {
				    IrpContext->Irp->IoStatus.Status = Status;
				    
				    Ext2CompleteRequest(
					    IrpContext->Irp,
					    (CCHAR)
					    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
					    );
				    
				    IrpContext->Irp = NULL;
			    }
			    
			    Ext2QueueCloseRequest(IrpContext);
		    }
		    else
		    {
			    if (IrpContext->Irp != NULL)
			    {
				    IrpContext->Irp->IoStatus.Status = Status;
				    
				    Ext2CompleteRequest(
					    IrpContext->Irp,
					    (CCHAR)
					    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
					    );
			    }
			    
			    Ext2FreeIrpContext(IrpContext);
			    
			    if (FreeVcb)
			    {
				    Ext2FreeVcb(Vcb);
			    }
		    }
	    }
    }
    
    return Status;
}

VOID
Ext2QueueCloseRequest (IN PEXT2_IRP_CONTEXT IrpContext)
{
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	// IsSynchronous means we can block (so we don't requeue it)
	IrpContext->IsSynchronous = TRUE;
	
	ExInitializeWorkItem(
		&IrpContext->WorkQueueItem,
		Ext2DeQueueCloseRequest,
		IrpContext);
	
	ExQueueWorkItem(&IrpContext->WorkQueueItem, CriticalWorkQueue);
}

VOID
Ext2DeQueueCloseRequest (IN PVOID Context)
{
	PEXT2_IRP_CONTEXT IrpContext;
	
	IrpContext = (PEXT2_IRP_CONTEXT) Context;
	
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	__try
	{
		__try
		{
			FsRtlEnterFileSystem();
			Ext2Close(IrpContext);
		}
		__except (Ext2ExceptionFilter(IrpContext, GetExceptionCode()))
		{
			Ext2ExceptionHandler(IrpContext);
		}
	}
	__finally
	{
		FsRtlExitFileSystem();
	}
}


NTSTATUS
Ext2LookupFileName (IN PEXT2_VCB        Vcb,
		    IN PUNICODE_STRING  FullFileName,
		    IN OUT PULONG       Offset,
		    IN OUT PULONG	Inode,
		    IN OUT PULONG	DirInode,
		    IN OUT PEXT2_INODE  ext2_inode)
{
	ULONG           RootInode = 2;
	UNICODE_STRING  FileName;
	NTSTATUS        Status = STATUS_OBJECT_NAME_NOT_FOUND;
	EXT2_DIR_ENTRY  ext2_dir;
	USHORT		i = 0;
	BOOLEAN		bRun = TRUE;
	BOOLEAN		bCurrent = FALSE;
	EXT2_INODE	in;
	ULONG		off = 0;

	*Offset = 0;
	*Inode = 0;

	if (*DirInode != 0 && FullFileName->Buffer[0] != L'\\')
	{
		bCurrent = TRUE;
		RootInode = *DirInode;
	}
	else
	{
		RootInode = 2;
		*DirInode = 2;
	}
	
	RtlZeroMemory(&ext2_dir, sizeof(EXT2_DIR_ENTRY));

	if (FullFileName->Length == 0)
	{
		return Status;
	}
	
	if (FullFileName->Length == 2 && FullFileName->Buffer[0] == L'\\')
	{
		if (!Ext2LoadInode(Vcb, RootInode, ext2_inode))
		{
			return Status;
		}
		*Inode = 2;
		*DirInode = 2;
		return STATUS_SUCCESS;
	}
	
	while (bRun && i < FullFileName->Length/2)
	{
		ULONG Length;

		if (bCurrent)
		{
			bCurrent = FALSE;
		}
		else
		{
			if(FullFileName->Buffer[i++] != L'\\')
				break;
		}

		Length = i;

		while(i < FullFileName->Length/2 && (FullFileName->Buffer[i] != L'\\'))
			i++;
		if (i - Length >0)
		{
			if (NT_SUCCESS(Status))
				RootInode = ext2_dir.inode;

			FileName = *FullFileName;
			FileName.Buffer += Length;
			FileName.Length = (USHORT)((i - Length) * 2);

			if (!Ext2LoadInode(Vcb, RootInode, &in))
			{
				return STATUS_OBJECT_NAME_NOT_FOUND;
			}

			if (!S_ISDIR(in.i_mode))
				break;

			Status = Ext2ScanDir (
				Vcb,
				RootInode,
				&FileName,
				&off,
				&in,
				&ext2_dir);

			if (!NT_SUCCESS(Status))
				bRun = FALSE;

		}
	}

	if (NT_SUCCESS(Status))
	{
		if (Inode)
			*Inode = ext2_dir.inode;
		if (DirInode)
			*DirInode = RootInode;
		
		if (Offset)
			*Offset = 0;
		if (ext2_inode)
			Ext2LoadInode(Vcb, ext2_dir.inode, ext2_inode);
	}

	return Status;
}

NTSTATUS
Ext2ScanDir (IN PEXT2_VCB        Vcb,
	     IN ULONG            inode,
	     IN PUNICODE_STRING  FileName,
	     IN OUT PULONG       Index,
	     IN PEXT2_INODE	 ext2_inode,
	     IN OUT PEXT2_DIR_ENTRY  ext2_dir)
{
	NTSTATUS                Status = STATUS_UNSUCCESSFUL;
	USHORT                  InodeFileNameLength;
	UNICODE_STRING          InodeFileName;

	PVOID			DirectoryContent = NULL;
	struct			ext2_dir_entry* pDir = NULL;
	ULONG			dwBytes;
	BOOLEAN			bFound = FALSE;
	

	InodeFileName.Buffer = NULL;
	
	__try
	{

		DirectoryContent = ExAllocatePool(NonPagedPool,
			ext2_inode->i_size);
		if (!DirectoryContent)
		{
			Status = STATUS_INSUFFICIENT_RESOURCES;
			__leave;
		}

		Ext2ReadInode(Vcb, ext2_inode, 0, DirectoryContent, ext2_inode->i_size);
		dwBytes = 0;

		pDir = (struct ext2_dir_entry *) ((PUCHAR)DirectoryContent + dwBytes);

		while (!bFound && dwBytes < ext2_inode->i_size && pDir->inode)
		{
			InodeFileNameLength = pDir->name_len & 0xff;
			
			InodeFileName.Length = InodeFileName.MaximumLength =
				InodeFileNameLength * 2;
			
			InodeFileName.Buffer = ExAllocatePool(
				NonPagedPool,
				InodeFileNameLength * 2 + 2);

			RtlZeroMemory(InodeFileName.Buffer, InodeFileNameLength * 2 + 2);
			
			Ext2CharToWchar(
				InodeFileName.Buffer,
				pDir->name,
				InodeFileNameLength );

			KdPrint(("Ext2ScanDir: file:%S to comp.\n", InodeFileName.Buffer));

			if (!RtlCompareUnicodeString(
				FileName,
				&InodeFileName,
				TRUE ))
			{
				bFound = TRUE;
				*Index = dwBytes;
				RtlCopyMemory(ext2_dir, pDir, pDir->rec_len > sizeof(EXT2_DIR_ENTRY)
					? sizeof(EXT2_DIR_ENTRY) : pDir->rec_len);
				Status = STATUS_SUCCESS;
				KdPrint(("Ext2ScanDir: Found: inode=%xh\n", pDir->inode));
			}
			
			if (InodeFileName.Buffer != NULL)
			{
				ExFreePool(InodeFileName.Buffer);
				InodeFileName.Buffer = NULL;
			}
			
			dwBytes +=pDir->rec_len;
			pDir = (struct ext2_dir_entry *) ((PUCHAR)DirectoryContent + dwBytes);

		}

		if (!bFound)
			Status = STATUS_NO_SUCH_FILE;
	}

	__finally
	{
		if (InodeFileName.Buffer != NULL)
		{
			ExFreePool(InodeFileName.Buffer);
		}

		if (DirectoryContent)
			ExFreePool(DirectoryContent);
	}
	
	return Status;
}

NTSTATUS
Ext2OpenFile(PEXT2_VCB Vcb, PIRP Irp)
{
	NTSTATUS			Status = STATUS_OBJECT_NAME_NOT_FOUND;
	PIO_STACK_LOCATION  io_stack;
	PEXT2_FCB           Fcb = NULL;
	PEXT2_FCB			pParentFcb = NULL;
	PEXT2_CCB           Ccb = NULL;
	ULONG               found_index = 0;
	struct ext2_inode*  ext2_inode;
	PLIST_ENTRY         Link;
	PEXT2_FCB           TmpFcb;
	BOOLEAN             VcbResourceAcquired = FALSE;
	BOOLEAN				bFound = FALSE;
	BOOLEAN				bDir = FALSE;
	ULONG				inode, dir_inode;
	UNICODE_STRING	    FileName;

	FileName.Buffer = NULL;
	io_stack = IoGetCurrentIrpStackLocation(Irp);

	__try
	{
		ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			TRUE );
		
		VcbResourceAcquired = TRUE;
		
		ext2_inode = ExAllocatePool(
			NonPagedPool,
			sizeof(struct ext2_inode) );
		if (!ext2_inode)
			__leave;

		FileName.MaximumLength = 0x400;

		if (io_stack->FileObject->RelatedFileObject)
		{
			pParentFcb = (PEXT2_FCB)(io_stack->FileObject->RelatedFileObject->FsContext);
		}

		FileName =	io_stack->FileObject->FileName;


		if (pParentFcb)
		{
			dir_inode = pParentFcb->inode;
		}
		else
		{
			dir_inode = 2;
		}

		KdPrint(("Ext2OpenFile: %S Opt: %xh.\n",
			FileName.Buffer,
			io_stack->Parameters.Create.Options));


		Status = Ext2LookupFileName(
			Vcb,
			&FileName,
			&found_index,
			&inode,
			&dir_inode,
			ext2_inode );

		if (!NT_SUCCESS(Status))
		{
			__leave;
		}
/*
		else
		{
			bDir = (io_stack->FileObject->Flags & 0x02) ? TRUE : FALSE;

			if (io_stack->FileObject->FileName.Buffer[0] == L'\\' &&
			    io_stack->FileObject->FileName.Length == 2 && 
			    Vcb->inode != 0 )
			    bDir = FALSE;

			if (S_ISDIR(ext2_inode->i_mode) && (Vcb->inode != inode) && bDir)
			{
				Vcb->inode = inode;
				if (Vcb->Name.Buffer)
					ExFreePool(Vcb->Name.Buffer);
				Vcb->Name.Buffer = ExAllocatePool(NonPagedPool, Length + 2);
				RtlZeroMemory(Vcb->Name.Buffer, Length + 2);
				Vcb->Name.MaximumLength = 0x400;
				Vcb->Name.Length = 0;
				RtlAppendUnicodeStringToString(
					&(Vcb->Name), &FileName);
				if (Vcb->Name.Buffer[(Length - 2)/2] != L'\\')
				{
					Vcb->Name.Buffer[Length/2] =  L'\\';
					Vcb->Name.Length +=2;
				}
			}
			else
			{
//				Vcb->inode = dir_inode;
			}

		}
*/

		Link = Vcb->FcbList.Flink;
		
		while (!bFound && Link != &Vcb->FcbList)
		{
			TmpFcb = CONTAINING_RECORD(Link, EXT2_FCB, Next);
			
			if (TmpFcb && TmpFcb->Identifier.Type == FCB)
			{
#if DBG
				KdPrint(("Ext2OpenFile: [%s,%xh]\n", TmpFcb->DbgFileName, TmpFcb->inode));
#endif				
				if (TmpFcb->inode == inode)
				{
					KdPrint(("Ext2OpenFile: Found FCB for %xh.\n", inode));
					Fcb = TmpFcb;
					bFound = TRUE;
				}
			}
			Link = Link->Flink;
		}

		
		if (!Fcb)
		{
			Fcb = Ext2AllocateFcb (
				Vcb, &FileName, found_index, inode,
				dir_inode, ext2_inode );
		}
		
		if (Fcb)
		{
			Ccb = Ext2AllocateCcb();
			
			Fcb->OpenHandleCount++;
			Vcb->OpenFileHandleCount++;
			Fcb->ReferenceCount++;
			Vcb->ReferenceCount++;
#if DBG
			KdPrint(("Ext2OpenFile: %s refercount=%xh\n", Fcb->DbgFileName, Fcb->ReferenceCount));
#endif			
			io_stack->FileObject->FsContext = (void*)Fcb;
			io_stack->FileObject->FsContext2 = (void*) Ccb;
			io_stack->FileObject->PrivateCacheMap = NULL;
			io_stack->FileObject->SectionObjectPointer = &(Fcb->SectionObject);
			io_stack->FileObject->Vpb = Vcb->Vpb;
			
			Irp->IoStatus.Information = FILE_OPENED;
			Status = STATUS_SUCCESS;
#if DBG			
			KdPrint(("Ext2OpenFile: %s OpenCount: %u ReferCount: %u\n",
				Fcb->DbgFileName, Fcb->OpenHandleCount,	Fcb->ReferenceCount));
#endif
		}
		else
		{
			Irp->IoStatus.Information = 0;
			Status = STATUS_OBJECT_NAME_NOT_FOUND;
			KdPrint(("Ext2OpenFile: Create denied (CCB allocate error.)\n"));
		}
	
	}

	__finally
	{
		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread() );
		}

		if (bFound)
		{
			if (ext2_inode)
				ExFreePool(ext2_inode);
		}
		else
		{
			if (!Fcb && ext2_inode)
				ExFreePool(ext2_inode);
		}
/*
		if (FileName.Buffer)
			ExFreePool(FileName.Buffer);
*/
	}
	
	return Status;
}


NTSTATUS
Ext2Create (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT      DeviceObject;
	PIRP                Irp;
	PIO_STACK_LOCATION  io_stack;
	PEXT2_VCB           Vcb = 0;
	NTSTATUS	    Status = STATUS_OBJECT_NAME_NOT_FOUND;


	DeviceObject = IrpContext->DeviceObject;

	Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
	
	Irp = IrpContext->Irp;
	
	io_stack = IoGetCurrentIrpStackLocation(Irp);
	
	if (DeviceObject == Ext2Global->DeviceObject)
	{
		KdPrint(("Ext2Create: Create on main device object.\n"));
		
		Irp->IoStatus.Information = FILE_OPENED;
		Status = STATUS_SUCCESS;
		IrpContext->Irp->IoStatus.Status = Status;
		
		Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
		
		Ext2FreeIrpContext(IrpContext);
		
		return Status;
		
	}
	else if (io_stack->FileObject->FileName.Length == 0)
	{
		io_stack->FileObject->FsContext = Vcb;
		
		ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			TRUE );
		
		Vcb->ReferenceCount++;
		
		ExReleaseResourceForThreadLite(
			&Vcb->MainResource,
			ExGetCurrentResourceThread() );
		
		Irp->IoStatus.Information = FILE_OPENED;
		Status = STATUS_SUCCESS;
//		KdPrint(("Ext2Create: Create on volume succeded.\n"));
		IrpContext->Irp->IoStatus.Status = Status;
		
		Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
		
		Ext2FreeIrpContext(IrpContext);
		
		return Status;
	}
	
//	KdPrint(("Ext2Create: FileName: (%S)\n", io_stack->FileObject->FileName.Buffer));

	__try
	{
		Status = Ext2OpenFile(Vcb, Irp);
	}

	
	__finally
	{

		if (!IrpContext->ExceptionInProgress)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(
				IrpContext->Irp,
				(CCHAR)
				(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
				);
			
			Ext2FreeIrpContext(IrpContext);
			
			if (Vcb &&
				Vcb->Flags & VCB_DISMOUNT_PENDING &&
				!Vcb->ReferenceCount )
			{
				Ext2FreeVcb(Vcb);
			}
		}
	}
	
	return Status;
}


NTSTATUS
Ext2ReadFileData (IN PEXT2_VCB		  Vcb,
		  IN ULONG                Index,
		  IN struct ext2_inode*   Inode,
		  IN PLARGE_INTEGER       Offset,
		  IN ULONG                Length,
		  IN OUT PVOID            Buffer)
{
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	
	PAGED_CODE();
	
	ASSERT(Vcb != NULL);
	ASSERT(Inode != NULL);
	ASSERT(Offset != NULL);
	ASSERT(Buffer != NULL);
	
	KdPrint(("Ext2ReadFileData: Index: %#x Offset: %I64u Length: %u\n",
		Index,
		Offset->QuadPart,
		Length));

	Ext2ReadInode(Vcb, Inode, (ULONG)Offset->QuadPart, Buffer, Length);
	
	return STATUS_SUCCESS;
}


NTSTATUS
Ext2Read (IN PEXT2_IRP_CONTEXT IrpContext)
{
	NTSTATUS Status;
	
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	if (IrpContext->MinorFunction & IRP_MN_COMPLETE)
	{
		Status = Ext2ReadComplete(IrpContext);
	}
	else
	{
		Status = Ext2ReadNormal(IrpContext);
	}
	
	return Status;
}

NTSTATUS
Ext2ReadNormal (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT      DeviceObject;
	NTSTATUS            Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB           Vcb;
	PFILE_OBJECT        FileObject;
	PEXT2_FCB           Fcb;
	PEXT2_CCB           Ccb;
	PIRP                Irp;
	PIO_STACK_LOCATION  IoStackLocation;
	ULONG               Length;
	ULONG               ReturnedLength;
	LARGE_INTEGER       ByteOffset;
	BOOLEAN             PagingIo;
	BOOLEAN             Nocache;
	BOOLEAN             SynchronousIo;
	BOOLEAN             VcbResourceAcquired = FALSE;
	BOOLEAN             FcbMainResourceAcquired = FALSE;
	BOOLEAN             FcbPagingIoResourceAcquired = FALSE;
	PUCHAR              Buffer;
	PDEVICE_OBJECT      DeviceToVerify;

	__try
	{
		ASSERT(IrpContext);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
	
		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		FileObject = IrpContext->FileObject;
		
		Fcb = (PEXT2_FCB) FileObject->FsContext;
		
		ASSERT(Fcb);
		
		if (Fcb->Identifier.Type == VCB)
		{
			Irp = IrpContext->Irp;
			
			IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
			
			Length = IoStackLocation->Parameters.Read.Length;
			ByteOffset = IoStackLocation->Parameters.Read.ByteOffset;
			
			PagingIo = (Irp->Flags & IRP_PAGING_IO ? TRUE : FALSE);
			Nocache = (Irp->Flags & IRP_NOCACHE ? TRUE : FALSE);
			SynchronousIo = (FileObject->Flags & FO_SYNCHRONOUS_IO ? TRUE : FALSE);
			
			if (Length == 0)
			{
				Irp->IoStatus.Information = 0;
				Status = STATUS_SUCCESS;
				__leave;
			}
			
			if (!Nocache)
			{
				Status = STATUS_INVALID_PARAMETER;
				__leave;
			}
			
			if (!ExAcquireResourceSharedLite(
				&Vcb->MainResource,
				IrpContext->IsSynchronous ))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			VcbResourceAcquired = TRUE;
			
			if (ByteOffset.QuadPart >=
#ifndef EXT2_RO
				Vcb->PartitionInformation.PartitionLength.QuadPart
#else
				(LONGLONG)(Vcb->ext2_super_block->s_blocks_count - Vcb->ext2_super_block->s_free_blocks_count) * Vcb->ext2_block
#endif
				)
			{
				Irp->IoStatus.Information = 0;
				Status = STATUS_END_OF_FILE;
				__leave;
			}
				
			if ((ByteOffset.QuadPart + Length) >
#ifndef EXT2_RO
					Vcb->PartitionInformation.PartitionLength.QuadPart
#else
					(LONGLONG)(Vcb->ext2_super_block->s_blocks_count - Vcb->ext2_super_block->s_free_blocks_count) * Vcb->ext2_block
#endif
					)
			{
				Length = (ULONG) (
#ifndef EXT2_RO
					Vcb->PartitionInformation.PartitionLength.QuadPart -
#else
					(LONGLONG)(Vcb->ext2_super_block->s_blocks_count - Vcb->ext2_super_block->s_free_blocks_count) * Vcb->ext2_block -
#endif
					ByteOffset.QuadPart);
	
				Length &= ~(SECTOR_SIZE - 1);
			}
			
			if (KeGetCurrentIrql() > PASSIVE_LEVEL)
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			Buffer = Ext2GetUserBuffer(Irp);
			
			if (Buffer == NULL)
			{
				Status = STATUS_INVALID_USER_BUFFER;
				__leave;
			}
			
			Status = Ext2ReadDisk(
				Vcb->TargetDeviceObject,
				(ULONG)ByteOffset.QuadPart / SECTOR_SIZE,
				(ULONG)ByteOffset.QuadPart % SECTOR_SIZE,
				Length, Buffer);
			
			if (Status == STATUS_VERIFY_REQUIRED)
			{
				Status = IoVerifyVolume(Vcb->TargetDeviceObject, FALSE);
				
				if (NT_SUCCESS(Status))
				{
					Status = Ext2ReadDisk(
						Vcb->TargetDeviceObject,
						(ULONG)ByteOffset.QuadPart / SECTOR_SIZE,
						(ULONG)ByteOffset.QuadPart % SECTOR_SIZE,
						Length, Buffer);
				}
			}
		
			if (NT_SUCCESS(Status))
			{
				Irp->IoStatus.Information = Length;
			}
			
			__leave;
		}
	
		ASSERT((Fcb->Identifier.Type == FCB) &&
			(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
		
		//
		// This request is not allowed on directories
		//
		if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			Status = STATUS_INVALID_PARAMETER;
			__leave;
		}
		
		Ccb = (PEXT2_CCB) FileObject->FsContext2;
		
		ASSERT(Ccb);
		
		ASSERT((Ccb->Identifier.Type == CCB) &&
			(Ccb->Identifier.Size == sizeof(EXT2_CCB)));
		
		Irp = IrpContext->Irp;
		
		IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
		
		Length = IoStackLocation->Parameters.Read.Length;
		ByteOffset = IoStackLocation->Parameters.Read.ByteOffset;
		
		PagingIo = (Irp->Flags & IRP_PAGING_IO ? TRUE : FALSE);
		Nocache = (Irp->Flags & IRP_NOCACHE ? TRUE : FALSE);
		SynchronousIo = (FileObject->Flags & FO_SYNCHRONOUS_IO ? TRUE : FALSE);
		
		if (Length == 0)
		{
			Irp->IoStatus.Information = 0;
			Status = STATUS_SUCCESS;
			__leave;
		}
		
		if (IrpContext->MinorFunction & IRP_MN_DPC)
		{
			IrpContext->MinorFunction &= ~IRP_MN_DPC;
			Status = STATUS_PENDING;
			__leave;
		}
		
		if (!PagingIo)
		{
			if (!ExAcquireResourceSharedLite(
				&Fcb->MainResource,
				IrpContext->IsSynchronous ))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			FcbMainResourceAcquired = TRUE;
		}
		else
		{
			if (!ExAcquireResourceSharedLite(
				&Fcb->PagingIoResource,
				IrpContext->IsSynchronous ))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			FcbPagingIoResourceAcquired = TRUE;
		}
		
		if ((ULONG)ByteOffset.QuadPart >= (Fcb->ext2_inode->i_size))
		{
			Irp->IoStatus.Information = 0;
			Status = STATUS_END_OF_FILE;
			__leave;
		}
		
		if (!PagingIo)
		{
			if (!FsRtlCheckLockForReadAccess(
				&Fcb->FileLockAnchor,
				Irp			))
			{
				Status = STATUS_FILE_LOCK_CONFLICT;
				__leave;
			}
		}
		
		if (!Nocache)
		{
			if ((ULONG)(ByteOffset.QuadPart + Length) >
				Fcb->ext2_inode->i_size)
			{
				Length =
					Fcb->ext2_inode->i_size - ByteOffset.LowPart;
			}
			
			if (FileObject->PrivateCacheMap == NULL)
			{
				CcInitializeCacheMap(
					FileObject,
					(PCC_FILE_SIZES)(&Fcb->CommonFCBHeader.AllocationSize),
					FALSE,
					&Ext2Global->CacheManagerCallbacks,
					Fcb );
			}
			
			if (IrpContext->MinorFunction & IRP_MN_MDL)
			{
				CcMdlRead(
					FileObject,
					&ByteOffset,
					Length,
					&Irp->MdlAddress,
					&Irp->IoStatus );
				
				Status = Irp->IoStatus.Status;
			}
			else
			{
				Buffer = Ext2GetUserBuffer(Irp);
				
				if (Buffer == NULL)
				{
					Status = STATUS_INVALID_USER_BUFFER;
					__leave;
				}
				
				if (!CcCopyRead(
					FileObject,
					&ByteOffset,
					Length,
					IrpContext->IsSynchronous,
					Buffer,
					&Irp->IoStatus ))
				{
					Status = STATUS_PENDING;
					__leave;
				}
				
				Status = Irp->IoStatus.Status;
			}
		}
		else
		{
			ReturnedLength = Length;
			
			if ((ByteOffset.QuadPart + Length) >
				Fcb->ext2_inode->i_size)
			{
				ReturnedLength =
					Fcb->ext2_inode->i_size - ByteOffset.LowPart;
				
				Length = (ReturnedLength & ~(SECTOR_SIZE - 1)) + SECTOR_SIZE;
			}
			
			if (KeGetCurrentIrql() > PASSIVE_LEVEL)
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			Buffer = Ext2GetUserBuffer(Irp);
			
			if (Buffer == NULL)
			{
				Status = STATUS_INVALID_USER_BUFFER;
				__leave;
			}
			
			Status = Ext2ReadFileData(
				Vcb,
				(ULONG) Fcb->IndexNumber.QuadPart,
				Fcb->ext2_inode,
				&ByteOffset,
				Length,
				Buffer);
			
			if (Status == STATUS_VERIFY_REQUIRED)
			{
				DeviceToVerify = IoGetDeviceToVerify(PsGetCurrentThread());
				
				IoSetDeviceToVerify(PsGetCurrentThread(), NULL);
				
				Status = IoVerifyVolume(DeviceToVerify, FALSE);
				
				if (NT_SUCCESS(Status))
				{
					Status = Ext2ReadFileData(
						Vcb,
						(ULONG) Fcb->IndexNumber.QuadPart,
						Fcb->ext2_inode,
						&ByteOffset,
						Length,
						Buffer );
				}
			}
			
			if (NT_SUCCESS(Status))
			{
				Irp->IoStatus.Information = ReturnedLength;
			}
		}
	}


	__finally
	{
		if (FcbPagingIoResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->PagingIoResource,
				ExGetCurrentResourceThread());
		}
		
		if (FcbMainResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread());
		}
		
		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread());
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			if (Status == STATUS_PENDING)
			{
				Status = Ext2LockUserBuffer(
					IrpContext->Irp,
					Length,
					IoWriteAccess );
				
				if (NT_SUCCESS(Status))
				{
					Status = Ext2QueueRequest(IrpContext);
				}
				else
				{
					IrpContext->Irp->IoStatus.Status = Status;
					Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
					Ext2FreeIrpContext(IrpContext);
				}
			}
			else
			{
				IrpContext->Irp->IoStatus.Status = Status;
				
				if (SynchronousIo && !PagingIo && NT_SUCCESS(Status))
				{
					FileObject->CurrentByteOffset.QuadPart =
						ByteOffset.QuadPart + Irp->IoStatus.Information;
				}
				
				if (!PagingIo && NT_SUCCESS(Status))
				{
					FileObject->Flags &= ~FO_FILE_FAST_IO_READ;
				}
				
				Ext2CompleteRequest(
					IrpContext->Irp,
					(CCHAR)
					(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
					);
				
				Ext2FreeIrpContext(IrpContext);
			}
		}
	}
	
	return Status;
}

NTSTATUS
Ext2ReadComplete (IN PEXT2_IRP_CONTEXT IrpContext)
{
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	PFILE_OBJECT    FileObject;
	PIRP            Irp;
	
	__try
	{
		ASSERT(IrpContext);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		FileObject = IrpContext->FileObject;
		
		Irp = IrpContext->Irp;
		
		CcMdlReadComplete(FileObject, Irp->MdlAddress);
		
		Irp->MdlAddress = NULL;
		
		Status = STATUS_SUCCESS;
	}
	__finally
	{
		if (!IrpContext->ExceptionInProgress)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(
				IrpContext->Irp,
				(CCHAR)
				(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
				);
			
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}

NTSTATUS
Ext2QueryInformation (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT          DeviceObject;
	NTSTATUS                Status = STATUS_UNSUCCESSFUL;
	PFILE_OBJECT            FileObject;
	PEXT2_FCB               Fcb;
	PEXT2_CCB               Ccb;
	PIRP                    Irp;
	PIO_STACK_LOCATION      IoStackLocation;
	FILE_INFORMATION_CLASS  FileInformationClass;
	ULONG                   Length;
	PVOID                   Buffer;
	BOOLEAN                 FcbResourceAcquired = FALSE;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		
		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		FileObject = IrpContext->FileObject;
		
		Fcb = (PEXT2_FCB) FileObject->FsContext;
		
		ASSERT(Fcb != NULL);
		
		//
		// This request is not allowed on volumes
		//
		if (Fcb->Identifier.Type == VCB)
		{
			Status = STATUS_INVALID_PARAMETER;
			__leave;
		}
		
		ASSERT((Fcb->Identifier.Type == FCB) &&
			(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
		
#ifndef EXT2_RO
		if (!Fcb->IsPageFile)
#endif
		{
			if (!ExAcquireResourceSharedLite(
				&Fcb->MainResource,
				IrpContext->IsSynchronous
				))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			FcbResourceAcquired = TRUE;
		}
		
		Ccb = (PEXT2_CCB) FileObject->FsContext2;
		
		ASSERT(Ccb != NULL);
		
		ASSERT((Ccb->Identifier.Type == CCB) &&
			(Ccb->Identifier.Size == sizeof(EXT2_CCB)));
		
		Irp = IrpContext->Irp;
		
		IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
		
		FileInformationClass =
			IoStackLocation->Parameters.QueryFile.FileInformationClass;
		
		Length = IoStackLocation->Parameters.QueryFile.Length;
		
		Buffer = Irp->AssociatedIrp.SystemBuffer;
		
		RtlZeroMemory(Buffer, Length);
		
		switch (FileInformationClass)
		{
		case FileBasicInformation:
			{
				PFILE_BASIC_INFORMATION FileBasicInformation;
				
				if (Length < sizeof(FILE_BASIC_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FileBasicInformation = (PFILE_BASIC_INFORMATION) Buffer;
				
				FileBasicInformation->CreationTime = Ext2SysTime(Fcb->ext2_inode->i_ctime + TIMEZONE);
				
				FileBasicInformation->LastAccessTime = Ext2SysTime(Fcb->ext2_inode->i_atime + TIMEZONE);
				
				FileBasicInformation->LastWriteTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
				
				FileBasicInformation->ChangeTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
				
				FileBasicInformation->FileAttributes = Fcb->FileAttributes;
				
				Irp->IoStatus.Information = sizeof(FILE_BASIC_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
		case FileAttributeTagInformation:
			{
				PFILE_ATTRIBUTE_TAG_INFORMATION FATI;
				
				if (Length < sizeof(FILE_ATTRIBUTE_TAG_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FATI = (PFILE_ATTRIBUTE_TAG_INFORMATION) Buffer;
				
				FATI->FileAttributes = Fcb->FileAttributes;
				FATI->ReparseTag = 0;
				
				Irp->IoStatus.Information = sizeof(FILE_ATTRIBUTE_TAG_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileStandardInformation:
			{
				PFILE_STANDARD_INFORMATION FileStandardInformation;
				
				if (Length < sizeof(FILE_STANDARD_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FileStandardInformation = (PFILE_STANDARD_INFORMATION) Buffer;
				
				FileStandardInformation->AllocationSize.QuadPart =
					Fcb->ext2_inode->i_size;
				
				FileStandardInformation->EndOfFile.QuadPart =
					Fcb->ext2_inode->i_size;
				
				FileStandardInformation->NumberOfLinks = Fcb->ext2_inode->i_links_count;
				
#ifndef EXT2_RO
				
				FileStandardInformation->DeletePending = Fcb->DeletePending;
				
#else
				
				FileStandardInformation->DeletePending = FALSE;
				
#endif
				
				if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				{
					FileStandardInformation->Directory = TRUE;
				}
				else
				{
					FileStandardInformation->Directory = FALSE;
				}
				
				Irp->IoStatus.Information = sizeof(FILE_STANDARD_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileInternalInformation:
			{
				PFILE_INTERNAL_INFORMATION FileInternalInformation;
				
				if (Length < sizeof(FILE_INTERNAL_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FileInternalInformation = (PFILE_INTERNAL_INFORMATION) Buffer;
				
				// The "inode number"
				FileInternalInformation->IndexNumber = Fcb->IndexNumber;
				
				Irp->IoStatus.Information = sizeof(FILE_INTERNAL_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileEaInformation:
			{
				PFILE_EA_INFORMATION FileEaInformation;
				
				if (Length < sizeof(FILE_EA_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FileEaInformation = (PFILE_EA_INFORMATION) Buffer;
				
				// Romfs doesn't have any extended attributes
				FileEaInformation->EaSize = 0;
				
				Irp->IoStatus.Information = sizeof(FILE_EA_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileNameInformation:
			{
				PFILE_NAME_INFORMATION FileNameInformation;
				
				if (Length < sizeof(FILE_NAME_INFORMATION) +
					Fcb->FileName.Length - sizeof(WCHAR))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FileNameInformation = (PFILE_NAME_INFORMATION) Buffer;
				
				FileNameInformation->FileNameLength = Fcb->FileName.Length;
				
				RtlCopyMemory(
					FileNameInformation->FileName,
					Fcb->FileName.Buffer,
					Fcb->FileName.Length
					);
				
				Irp->IoStatus.Information = sizeof(FILE_NAME_INFORMATION) +
					Fcb->FileName.Length - sizeof(WCHAR);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FilePositionInformation:
			{
				PFILE_POSITION_INFORMATION FilePositionInformation;
				
				if (Length < sizeof(FILE_POSITION_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FilePositionInformation = (PFILE_POSITION_INFORMATION) Buffer;
				
				FilePositionInformation->CurrentByteOffset =
					FileObject->CurrentByteOffset;
				
				Irp->IoStatus.Information = sizeof(FILE_POSITION_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileAllInformation:
			{
				PFILE_ALL_INFORMATION       FileAllInformation;
				PFILE_BASIC_INFORMATION     FileBasicInformation;
				PFILE_STANDARD_INFORMATION  FileStandardInformation;
				PFILE_INTERNAL_INFORMATION  FileInternalInformation;
				PFILE_EA_INFORMATION        FileEaInformation;
				PFILE_POSITION_INFORMATION  FilePositionInformation;
				PFILE_NAME_INFORMATION      FileNameInformation;
				
				if (Length < sizeof(FILE_ALL_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FileAllInformation = (PFILE_ALL_INFORMATION) Buffer;
				
				FileBasicInformation =
					&FileAllInformation->BasicInformation;
				
				FileStandardInformation =
					&FileAllInformation->StandardInformation;
				
				FileInternalInformation =
					&FileAllInformation->InternalInformation;
				
				FileEaInformation =
					&FileAllInformation->EaInformation;
				
				FilePositionInformation =
					&FileAllInformation->PositionInformation;
				
				FileNameInformation =
					&FileAllInformation->NameInformation;
				
				FileBasicInformation->CreationTime = Ext2SysTime(Fcb->ext2_inode->i_ctime + TIMEZONE);
				
				FileBasicInformation->LastAccessTime = Ext2SysTime(Fcb->ext2_inode->i_atime + TIMEZONE);
				
				FileBasicInformation->LastWriteTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
				
				FileBasicInformation->ChangeTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
				
				FileBasicInformation->FileAttributes = Fcb->FileAttributes;
				
				FileStandardInformation->AllocationSize.QuadPart =
					Fcb->ext2_inode->i_size;
				
				FileStandardInformation->EndOfFile.QuadPart =
					Fcb->ext2_inode->i_size;
				
				FileStandardInformation->NumberOfLinks = Fcb->ext2_inode->i_links_count;
#ifndef EXT2_RO
				FileStandardInformation->DeletePending = Fcb->DeletePending;
#else
				FileStandardInformation->DeletePending = FALSE;
#endif
				
				if (Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				{
					FileStandardInformation->Directory = TRUE;
				}
				else
				{
					FileStandardInformation->Directory = FALSE;
				}
				
				// The "inode number"
				FileInternalInformation->IndexNumber = Fcb->IndexNumber;
				
				// Romfs doesn't have any extended attributes
				FileEaInformation->EaSize = 0;
				
				FilePositionInformation->CurrentByteOffset =
					FileObject->CurrentByteOffset;
				
				if (Length < sizeof(FILE_ALL_INFORMATION) +
					Fcb->FileName.Length - sizeof(WCHAR))
				{
					Irp->IoStatus.Information = sizeof(FILE_ALL_INFORMATION);
					Status = STATUS_BUFFER_OVERFLOW;
					__leave;
				}
				
				FileNameInformation->FileNameLength = Fcb->FileName.Length;
				
				RtlCopyMemory(
					FileNameInformation->FileName,
					Fcb->FileName.Buffer,
					Fcb->FileName.Length
					);
				
				Irp->IoStatus.Information = sizeof(FILE_ALL_INFORMATION) +
					Fcb->FileName.Length - sizeof(WCHAR);
				Status = STATUS_SUCCESS;
				__leave;
            }
	    
	    /*
	    case FileAlternateNameInformation:
            {
	    // TODO: Handle FileAlternateNameInformation
	    
	      // Here we would like to use RtlGenerate8dot3Name but I don't
	      // know how to use the argument PGENERATE_NAME_CONTEXT
	      }
	      */
	      
        case FileNetworkOpenInformation:
		{
			PFILE_NETWORK_OPEN_INFORMATION FileNetworkOpenInformation;
			
			if (Length < sizeof(FILE_NETWORK_OPEN_INFORMATION))
			{
				Status = STATUS_INFO_LENGTH_MISMATCH;
				__leave;
			}
			
			FileNetworkOpenInformation =
				(PFILE_NETWORK_OPEN_INFORMATION) Buffer;
			
			FileNetworkOpenInformation->CreationTime = Ext2SysTime(Fcb->ext2_inode->i_ctime + TIMEZONE);
			
			FileNetworkOpenInformation->LastAccessTime = Ext2SysTime(Fcb->ext2_inode->i_atime + TIMEZONE);
			
			FileNetworkOpenInformation->LastWriteTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
			
			FileNetworkOpenInformation->ChangeTime = Ext2SysTime(Fcb->ext2_inode->i_mtime + TIMEZONE);
			
			FileNetworkOpenInformation->AllocationSize.QuadPart =
				Fcb->ext2_inode->i_size;
			
			FileNetworkOpenInformation->EndOfFile.QuadPart =
				Fcb->ext2_inode->i_size;
			
			FileNetworkOpenInformation->FileAttributes =
				Fcb->FileAttributes;
			
			Irp->IoStatus.Information =
				sizeof(FILE_NETWORK_OPEN_INFORMATION);
			Status = STATUS_SUCCESS;
			__leave;
		}
		
        default:
		Status = STATUS_INVALID_INFO_CLASS;
        }
    }
    __finally
    {
	    if (FcbResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Fcb->MainResource,
			    ExGetCurrentResourceThread());
	    }
	    
	    if (!IrpContext->ExceptionInProgress)
	    {
		    if (Status == STATUS_PENDING)
		    {
			    Ext2QueueRequest(IrpContext);
		    }
		    else
		    {
			    IrpContext->Irp->IoStatus.Status = Status;
			    
			    Ext2CompleteRequest(
				    IrpContext->Irp,
				    (CCHAR)
				    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
				    );
			    
			    Ext2FreeIrpContext(IrpContext);
		    }
	    }
    }
    
    return Status;
}


NTSTATUS
Ext2SetInformation (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT          DeviceObject;
	NTSTATUS                Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB               Vcb;
	PFILE_OBJECT            FileObject;
	PEXT2_FCB               Fcb;
	PEXT2_CCB               Ccb;
	PIRP                    Irp;
	PIO_STACK_LOCATION      IoStackLocation;
	FILE_INFORMATION_CLASS  FileInformationClass;
	ULONG                   Length;
	PVOID                   Buffer;
#ifndef EXT2_RO
	BOOLEAN                 VcbResourceAcquired = FALSE;
#endif
	BOOLEAN                 FcbMainResourceAcquired = FALSE;
#ifndef EXT2_RO
	BOOLEAN                 FcbPagingIoResourceAcquired = FALSE;
#endif

	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;

		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		FileObject = IrpContext->FileObject;
		
		Fcb = (PEXT2_FCB) FileObject->FsContext;
		
		ASSERT(Fcb != NULL);
		
		//
		// This request is not allowed on volumes
		//
		if (Fcb->Identifier.Type == VCB)
		{
			Status = STATUS_INVALID_PARAMETER;
			__leave;
		}
		
		ASSERT((Fcb->Identifier.Type == FCB) &&
			(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
		
		Ccb = (PEXT2_CCB) FileObject->FsContext2;
		
		ASSERT(Ccb != NULL);
		
		ASSERT((Ccb->Identifier.Type == CCB) &&
			(Ccb->Identifier.Size == sizeof(EXT2_CCB)));
		
		Irp = IrpContext->Irp;
		
		IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
		
		FileInformationClass =
			IoStackLocation->Parameters.SetFile.FileInformationClass;
		
		Length = IoStackLocation->Parameters.SetFile.Length;
		
		Buffer = Irp->AssociatedIrp.SystemBuffer;
		
#ifndef EXT2_RO
		
		if (FileInformationClass == FileDispositionInformation ||
			FileInformationClass == FileRenameInformation ||
			FileInformationClass == FileLinkInformation)
		{
			if (!ExAcquireResourceExclusiveLite(
				&Vcb->MainResource,
				IrpContext->IsSynchronous ))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			VcbResourceAcquired = TRUE;
		}
		
#endif // !EXT2_RO
		
#ifndef EXT2_RO
		if (!Fcb->IsPageFile)
#endif
		{
			if (!ExAcquireResourceExclusiveLite(
				&Fcb->MainResource,
				IrpContext->IsSynchronous ))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			FcbMainResourceAcquired = TRUE;
		}
		
#ifndef EXT2_RO
		
		if (FileInformationClass == FileDispositionInformation ||
			FileInformationClass == FileRenameInformation ||
			FileInformationClass == FileLinkInformation ||
			FileInformationClass == FileAllocationInformation ||
			FileInformationClass == FileEndOfFileInformation)
		{
			if (!ExAcquireResourceExclusiveLite(
				&Fcb->PagingIoResource,
				IrpContext->IsSynchronous ))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			FcbPagingIoResourceAcquired = TRUE;
		}
		
#endif // !EXT2_RO
		
#ifndef EXT2_RO
		if (Vcb->ReadOnly)
#endif
		{
			if (FileInformationClass != FilePositionInformation)
			{
				Status = STATUS_MEDIA_WRITE_PROTECTED;
				__leave;
			}
		}
		
		switch (FileInformationClass)
		{
			//
			// This is the only set file information request supported on read
			// only file systems
			//
		case FilePositionInformation:
			{
				PFILE_POSITION_INFORMATION FilePositionInformation;
				
				if (Length < sizeof(FILE_POSITION_INFORMATION))
				{
					Status = STATUS_INVALID_PARAMETER;
					__leave;
				}
				
				FilePositionInformation = (PFILE_POSITION_INFORMATION) Buffer;
				
				if ((FileObject->Flags & FO_NO_INTERMEDIATE_BUFFERING) &&
					(FilePositionInformation->CurrentByteOffset.LowPart &
					DeviceObject->AlignmentRequirement) )
				{
					Status = STATUS_INVALID_PARAMETER;
					__leave;
				}
				
				FileObject->CurrentByteOffset =
					FilePositionInformation->CurrentByteOffset;
				
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		default:
			Status = STATUS_INVALID_INFO_CLASS;
		}
    }
    __finally
    {
	    
#ifndef EXT2_RO
	    
	    if (FcbPagingIoResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Fcb->PagingIoResource,
			    ExGetCurrentResourceThread() );
	    }
	    
#endif // !EXT2_RO
	    
	    if (FcbMainResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Fcb->MainResource,
			    ExGetCurrentResourceThread() );
	    }
	    
#ifndef EXT2_RO
	    
	    if (VcbResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Vcb->MainResource,
			    ExGetCurrentResourceThread() );
	    }
	    
#endif // !EXT2_RO
	    
	    if (!IrpContext->ExceptionInProgress)
	    {
		    if (Status == STATUS_PENDING)
		    {
			    Ext2QueueRequest(IrpContext);
		    }
		    else
		    {
			    IrpContext->Irp->IoStatus.Status = Status;
			    
			    Ext2CompleteRequest(
				    IrpContext->Irp,
				    (CCHAR)
				    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
			    
			    Ext2FreeIrpContext(IrpContext);
		    }
	    }
    }
    
    return Status;
}


NTSTATUS
Ext2QueryVolumeInformation (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT          DeviceObject;
	NTSTATUS                Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB               Vcb;
	PIRP                    Irp;
	PIO_STACK_LOCATION      IoStackLocation;
	FS_INFORMATION_CLASS    FsInformationClass;
	ULONG                   Length;
	PVOID                   Buffer;
	BOOLEAN                 VcbResourceAcquired = FALSE;

	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
	
		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		if (!ExAcquireResourceSharedLite(
			&Vcb->MainResource,
			IrpContext->IsSynchronous
			))
		{
			Status = STATUS_PENDING;
			__leave;
		}
		
		VcbResourceAcquired = TRUE;
		
		Irp = IrpContext->Irp;
		
		IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
		
		FsInformationClass =
			IoStackLocation->Parameters.QueryVolume.FsInformationClass;
		
		Length = IoStackLocation->Parameters.QueryVolume.Length;
		
		Buffer = Irp->AssociatedIrp.SystemBuffer;
		
		RtlZeroMemory(Buffer, Length);
		
		switch (FsInformationClass)
		{
		case FileFsVolumeInformation:
			{
				PFILE_FS_VOLUME_INFORMATION FsVolInfo;
				ULONG                       VolumeLabelLength;
				ULONG                       RequiredLength;
				
				if (Length < sizeof(FILE_FS_VOLUME_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FsVolInfo = (PFILE_FS_VOLUME_INFORMATION) Buffer;
				
				FsVolInfo->VolumeCreationTime.QuadPart = 0;
				
				FsVolInfo->VolumeSerialNumber = 0;

				VolumeLabelLength = 1;
				
				FsVolInfo->VolumeLabelLength = VolumeLabelLength * 2;
				
				// I don't know what this means
				FsVolInfo->SupportsObjects = FALSE;
				
				RequiredLength = sizeof(FILE_FS_VOLUME_INFORMATION)
					+ VolumeLabelLength * 2 - sizeof(WCHAR);
				
				if (Length < RequiredLength)
				{
					Irp->IoStatus.Information =
						sizeof(FILE_FS_VOLUME_INFORMATION);
					Status = STATUS_BUFFER_OVERFLOW;
					__leave;
				}
				
				Ext2CharToWchar(
					FsVolInfo->VolumeLabel,
					"\\",
					VolumeLabelLength );
				
				Irp->IoStatus.Information = RequiredLength;
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileFsSizeInformation:
			{
				PFILE_FS_SIZE_INFORMATION FsSizeInfo;
				
				if (Length < sizeof(FILE_FS_SIZE_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FsSizeInfo = (PFILE_FS_SIZE_INFORMATION) Buffer;
				
				{
					FsSizeInfo->TotalAllocationUnits.QuadPart =
						Vcb->ext2_super_block->s_blocks_count;
					
					FsSizeInfo->AvailableAllocationUnits.QuadPart = 0;
				}
				
				FsSizeInfo->SectorsPerAllocationUnit =
					Vcb->ext2_block / Vcb->DiskGeometry.BytesPerSector;
				
				FsSizeInfo->BytesPerSector =
					Vcb->DiskGeometry.BytesPerSector;
				
				Irp->IoStatus.Information = sizeof(FILE_FS_SIZE_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileFsDeviceInformation:
			{
				PFILE_FS_DEVICE_INFORMATION FsDevInfo;
				
				if (Length < sizeof(FILE_FS_DEVICE_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FsDevInfo = (PFILE_FS_DEVICE_INFORMATION) Buffer;
				
				FsDevInfo->DeviceType =
					Vcb->TargetDeviceObject->DeviceType;
				
				FsDevInfo->Characteristics =
					Vcb->TargetDeviceObject->Characteristics;
				
#ifndef EXT2_RO
				if (Vcb->ReadOnly)
#endif
				{
					FsDevInfo->Characteristics |=
						FILE_READ_ONLY_DEVICE;
				}
				
				Irp->IoStatus.Information = sizeof(FILE_FS_DEVICE_INFORMATION);
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		case FileFsAttributeInformation:
			{
				PFILE_FS_ATTRIBUTE_INFORMATION  FsAttrInfo;
				ULONG                           RequiredLength;
				
				if (Length < sizeof(FILE_FS_ATTRIBUTE_INFORMATION))
				{
					Status = STATUS_INFO_LENGTH_MISMATCH;
					__leave;
				}
				
				FsAttrInfo =
					(PFILE_FS_ATTRIBUTE_INFORMATION) Buffer;
				
				FsAttrInfo->FileSystemAttributes =
					FILE_CASE_SENSITIVE_SEARCH | FILE_CASE_PRESERVED_NAMES;
				
				FsAttrInfo->MaximumComponentNameLength = EXT2_NAME_LEN;
				
				FsAttrInfo->FileSystemNameLength = sizeof(DRIVER_NAME) * 2;
				
				RequiredLength = sizeof(FILE_FS_ATTRIBUTE_INFORMATION) +
					sizeof(DRIVER_NAME) * 2 - sizeof(WCHAR);
				
				if (Length < RequiredLength)
				{
					Irp->IoStatus.Information =
						sizeof(FILE_FS_ATTRIBUTE_INFORMATION);
					Status = STATUS_BUFFER_OVERFLOW;
					__leave;
				}
				
				Ext2CharToWchar(
					FsAttrInfo->FileSystemName,
					DRIVER_NAME,
					sizeof(DRIVER_NAME));
				
				Irp->IoStatus.Information = RequiredLength;
				Status = STATUS_SUCCESS;
				__leave;
			}
			
		default:
			Status = STATUS_INVALID_INFO_CLASS;
        }
    }
    __finally
    {
	    if (VcbResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Vcb->MainResource,
			    ExGetCurrentResourceThread() );
	    }
	    
	    if (!IrpContext->ExceptionInProgress)
	    {
		    if (Status == STATUS_PENDING)
		    {
			    Ext2QueueRequest(IrpContext);
		    }
		    else
		    {
			    IrpContext->Irp->IoStatus.Status = Status;
			    
			    Ext2CompleteRequest(
				    IrpContext->Irp,
				    (CCHAR)
				    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
			    
			    Ext2FreeIrpContext(IrpContext);
		    }
	    }
    }
    
    return Status;
}


VOID
Ext2CharToWchar (IN OUT PWCHAR   Destination,
		 IN PCHAR        Source,
		 IN ULONG        Length)
{
	ULONG Index;
	
	ASSERT(Destination != NULL);
	ASSERT(Source != NULL);
	
	for (Index = 0; Index < Length; Index++)
	{
		Destination[Index] = (WCHAR) Source[Index];
	}
}

NTSTATUS
Ext2WcharToChar (IN OUT PCHAR    Destination,
		 IN PWCHAR       Source,
		 IN ULONG        Length)
{
	ULONG       Index;
	NTSTATUS    Status;
	
	ASSERT(Destination != NULL);
	ASSERT(Source != NULL);
	
	for (Index = 0, Status = STATUS_SUCCESS; Index < Length; Index++)
	{
		Destination[Index] = (CHAR) Source[Index];
		
		//
		// Check that the wide character fits in a normal character
		// but continue with the conversion anyway
		//
		if ( ((WCHAR) Destination[Index]) != Source[Index] )
		{
			Status = STATUS_OBJECT_NAME_INVALID;
		}
	}
	
	return Status;
}


NTSTATUS
Ext2LockUserBuffer (IN PIRP             Irp,
		    IN ULONG            Length,
		    IN LOCK_OPERATION   Operation)
{
	NTSTATUS Status;
	
	ASSERT(Irp != NULL);
	
	if (Irp->MdlAddress != NULL)
	{
		return STATUS_SUCCESS;
	}
	
	IoAllocateMdl(Irp->UserBuffer, Length, FALSE, FALSE, Irp);
	
	if (Irp->MdlAddress == NULL)
	{
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	
	__try
	{
		MmProbeAndLockPages(Irp->MdlAddress, Irp->RequestorMode, Operation);
		
		Status = STATUS_SUCCESS;
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
		IoFreeMdl(Irp->MdlAddress);
		
		Irp->MdlAddress = NULL;
		
		Status = STATUS_INVALID_USER_BUFFER;
	}
	
	return Status;
}

PVOID
Ext2GetUserBuffer (IN PIRP Irp )
{
	ASSERT(Irp != NULL);
	
	if (Irp->MdlAddress != NULL)
	{
		return MmGetSystemAddressForMdl(Irp->MdlAddress);
	}
	else
	{
		return Irp->UserBuffer;
	}
}

ULONG
Ext2GetInfoLength(IN FILE_INFORMATION_CLASS  FileInformationClass)
{
	switch (FileInformationClass)
	{
	case FileDirectoryInformation:
		return sizeof(FILE_DIRECTORY_INFORMATION);
		break;
		
	case FileFullDirectoryInformation:
		return sizeof(FILE_FULL_DIR_INFORMATION);
		break;
		
	case FileBothDirectoryInformation:
		return sizeof(FILE_BOTH_DIR_INFORMATION);
		break;
		
	case FileNamesInformation:
		return sizeof(FILE_NAMES_INFORMATION);
		break;
		
	default:
		break;
	}

	return 0;
}

/*
#define FillInfo (FI, BSize, Inode, Index, NSize, pName, Single) {\
	if (!Single) \
		FI->NextEntryOffset = BSize + NSize - sizeof(WCHAR); \
	else \
			FI->NextEntryOffset = 0; \
	FI->FileIndex = Index; \
	FI->CreationTime.QuadPart = Inode.i_ctime + TIMEZONE; \
	FI->LastAccessTime.QuadPart = Inode.i_atime + TIMEZONE; \
	FI->LastWriteTime.QuadPart = Inode.i_mtime + TIMEZONE; \
	FI->ChangeTime.QuadPart = Inode.i_mtime + TIMEZONE; \
	FI->EndOfFile.QuadPart = Inode.i_size; \
	FI->AllocationSize.QuadPart = Inode.i_size; \
	FI->LastAccessTime.QuadPart = Inode.i_atime + TIMEZONE; \
	FI->FileAttributes = FILE_ATTRIBUTE_NORMAL; \
	if (S_ISDIR(Inode->i_mode)) \
		FI->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY; \
	FI->FileNameLength = NSize; \
	RtlCopyMemory(FI->FileName, pName->Buffer, NSize); \
	dwBytes = BSize + NSize - sizeof(WCHAR); }
*/

ULONG
Ext2ProcessDirEntry(IN PEXT2_VCB         Vcb,
		    IN FILE_INFORMATION_CLASS  FileInformationClass,
		    IN ULONG		 in,
		    IN PVOID		 Buffer,
		    IN ULONG		 UsedLength,
		    IN ULONG		 Length,
		    IN ULONG		 FileIndex,
		    IN PUNICODE_STRING	 pName,
		    IN BOOLEAN		 Single )
{
	EXT2_INODE inode;
	PFILE_DIRECTORY_INFORMATION FDI;
	PFILE_FULL_DIR_INFORMATION FFI;
	PFILE_BOTH_DIR_INFORMATION FBI;
	PFILE_NAMES_INFORMATION	FNI;

	ULONG InfoLength = 0;
	ULONG NameLength = 0;
	ULONG dwBytes = 0;

	NameLength = pName->Length;
//	KdPrint(("Ext2ProcessDirEntry: NameLength: %xh.\n", NameLength));

	if (!in)
	{
		KdPrint(("Ext2PricessDirEntry: ext2_dir_entry is empty.\n"));
		return 0;
	}

	InfoLength = Ext2GetInfoLength(FileInformationClass);
	if (!InfoLength || InfoLength + NameLength - sizeof(WCHAR)> Length)
	{
		KdPrint(("Ext2PricessDirEntry: Size/Length error.\n"));
		return 0;
	}

	if(!Ext2LoadInode(Vcb, in, &inode))
	{
		KdPrint(("Ext2PricessDirEntry: Loading inode %xh error.\n", in));
		return 0;
	}

	switch(FileInformationClass)
	{
	case FileDirectoryInformation:
		FDI = (PFILE_DIRECTORY_INFORMATION) ((PUCHAR)Buffer + UsedLength);
		//FillInfo (FDI, InfoLength, inode, FileIndex, NameLength, pName, Single)
		if (!Single)
			FDI->NextEntryOffset = InfoLength + NameLength - sizeof(WCHAR);
		else
			FDI->NextEntryOffset = 0;
		FDI->FileIndex = FileIndex;
		FDI->CreationTime = Ext2SysTime(inode.i_ctime + TIMEZONE);
		FDI->LastAccessTime = Ext2SysTime(inode.i_atime + TIMEZONE);
		FDI->LastWriteTime = Ext2SysTime(inode.i_mtime + TIMEZONE);
		FDI->ChangeTime = Ext2SysTime(inode.i_mtime + TIMEZONE);
		FDI->EndOfFile.QuadPart = inode.i_size;
		FDI->AllocationSize.QuadPart = inode.i_size;
		FDI->FileAttributes = FILE_ATTRIBUTE_NORMAL;
		if (S_ISDIR(inode.i_mode))
			FDI->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
		FDI->FileNameLength = NameLength;
		RtlCopyMemory(FDI->FileName, pName->Buffer, NameLength);
		dwBytes = InfoLength + NameLength - sizeof(WCHAR); 
		break;
		
	case FileFullDirectoryInformation:
		FFI = (PFILE_FULL_DIR_INFORMATION) ((PUCHAR)Buffer + UsedLength);
//		FillInfo (FFI, InfoLength, inode, FileIndex, NameLength, pName, Single)
		if (!Single)
			FFI->NextEntryOffset = InfoLength + NameLength - sizeof(WCHAR);
		else
			FFI->NextEntryOffset = 0;
		FFI->FileIndex = FileIndex;
		FFI->CreationTime = Ext2SysTime(inode.i_ctime + TIMEZONE);
		FFI->LastAccessTime = Ext2SysTime(inode.i_atime + TIMEZONE);
		FFI->LastWriteTime = Ext2SysTime(inode.i_mtime + TIMEZONE);
		FFI->ChangeTime = Ext2SysTime(inode.i_mtime + TIMEZONE);
		FFI->EndOfFile.QuadPart = inode.i_size;
		FFI->AllocationSize.QuadPart = inode.i_size;
		FFI->FileAttributes = FILE_ATTRIBUTE_NORMAL;
		if (S_ISDIR(inode.i_mode))
			FFI->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
		FFI->FileNameLength = NameLength;
		RtlCopyMemory(FFI->FileName, pName->Buffer, NameLength);
		dwBytes = InfoLength + NameLength - sizeof(WCHAR); 

		break;
		
	case FileBothDirectoryInformation:
		FBI = (PFILE_BOTH_DIR_INFORMATION) ((PUCHAR)Buffer + UsedLength);
//		FillInfo (FBI, InfoLength, inode, FileIndex, NameLength, pName, Single)
		if (!Single)
			FBI->NextEntryOffset = InfoLength + NameLength - sizeof(WCHAR);
		else
			FBI->NextEntryOffset = 0;
		FBI->CreationTime = Ext2SysTime(inode.i_ctime + TIMEZONE);
		FBI->LastAccessTime = Ext2SysTime(inode.i_atime + TIMEZONE);
		FBI->LastWriteTime = Ext2SysTime(inode.i_mtime + TIMEZONE);
		FBI->ChangeTime = Ext2SysTime(inode.i_mtime + TIMEZONE);

/*
		FBI->CreationTime.QuadPart = inode.i_ctime + TIMEZONE;
		FBI->LastAccessTime.QuadPart = inode.i_atime + TIMEZONE;
		FBI->LastWriteTime.QuadPart = inode.i_mtime + TIMEZONE;
		FBI->ChangeTime.QuadPart = inode.i_mtime + TIMEZONE;
*/
		FBI->EndOfFile.QuadPart = inode.i_size;
		FBI->AllocationSize.QuadPart = inode.i_size;
		FBI->FileAttributes = FILE_ATTRIBUTE_NORMAL;
		if (S_ISDIR(inode.i_mode))
			FBI->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
		FBI->FileNameLength = NameLength;
		RtlCopyMemory(FBI->FileName, pName->Buffer, NameLength);
		dwBytes = InfoLength + NameLength - sizeof(WCHAR); 

		break;
		
	case FileNamesInformation:
		FNI = (PFILE_NAMES_INFORMATION) ((PUCHAR)Buffer + UsedLength);
//		FillInfo (FNI, InfoLength, inode, FileIndex, NameLength, pName, Single)
		if (!Single)
			FNI->NextEntryOffset = InfoLength + NameLength - sizeof(WCHAR);
		else
			FNI->NextEntryOffset = 0;
		FNI->FileNameLength = NameLength;
		RtlCopyMemory(FNI->FileName, pName->Buffer, NameLength);
		dwBytes = InfoLength + NameLength - sizeof(WCHAR); 

		break;
		
	default:
		break;
	}

	return dwBytes;
}


NTSTATUS
Ext2QueryDirectory (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT          DeviceObject;
	NTSTATUS                Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB               Vcb;
	PFILE_OBJECT            FileObject;
	PEXT2_FCB               Fcb;
	PEXT2_CCB               Ccb;
	PIRP                    Irp;
	PIO_STACK_LOCATION      IoStackLocation;
	FILE_INFORMATION_CLASS  FileInformationClass;
	ULONG                   Length;
	PUNICODE_STRING         FileName;
	ULONG                   FileIndex;
	BOOLEAN                 RestartScan;
	BOOLEAN                 ReturnSingleEntry;
	BOOLEAN                 IndexSpecified;
	PUCHAR                  Buffer;
	BOOLEAN                 FirstQuery;
	struct ext2_inode*      Inode = NULL;
	BOOLEAN                 FcbResourceAcquired = FALSE;
	ULONG                   UsedLength = 0;
	USHORT                  InodeFileNameLength;
	UNICODE_STRING          InodeFileName;
	PVOID			DirectoryContent = NULL;
	struct			ext2_dir_entry* pDir;
	ULONG			dwBytes;
	ULONG			dwTemp = 0;
	ULONG			dwSize = 0;
	ULONG			dwReturn = 0;
	BOOLEAN			bRun = TRUE;

	InodeFileName.Buffer = NULL;
	
	__try
	{
		ASSERT(IrpContext);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		
		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		FileObject = IrpContext->FileObject;
		
		Fcb = (PEXT2_FCB) FileObject->FsContext;
		
		ASSERT(Fcb);
		
		//
		// This request is not allowed on volumes
		//
		if (Fcb->Identifier.Type == VCB)
		{
			Status = STATUS_INVALID_PARAMETER;
			__leave;
		}
		
		ASSERT((Fcb->Identifier.Type == FCB) &&
			(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
		
		if (!ExAcquireResourceSharedLite(
			&Fcb->MainResource,
			IrpContext->IsSynchronous
			))
		{
			Status = STATUS_PENDING;
			__leave;
		}
		
		FcbResourceAcquired = TRUE;
		
		//
		// This request is only allowed on directories
		//
		if (!(Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY))
		{
			Status = STATUS_INVALID_PARAMETER;
			__leave;
		}
		
		Ccb = (PEXT2_CCB) FileObject->FsContext2;
		
		ASSERT(Ccb);
		
		ASSERT((Ccb->Identifier.Type == CCB) &&
			(Ccb->Identifier.Size == sizeof(EXT2_CCB)));
		
		Irp = IrpContext->Irp;
		
		IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
		
#ifndef _GNU_NTIFS_
		
		FileInformationClass =
			IoStackLocation->Parameters.QueryDirectory.FileInformationClass;
		
		Length = IoStackLocation->Parameters.QueryDirectory.Length;
		
		FileName = IoStackLocation->Parameters.QueryDirectory.FileName;
		
		FileIndex = IoStackLocation->Parameters.QueryDirectory.FileIndex;
		
#else // _GNU_NTIFS_
		
		FileInformationClass = ((PEXTENDED_IO_STACK_LOCATION)
			IoStackLocation)->Parameters.QueryDirectory.FileInformationClass;
		
		Length = ((PEXTENDED_IO_STACK_LOCATION)
			IoStackLocation)->Parameters.QueryDirectory.Length;
		
		FileName = ((PEXTENDED_IO_STACK_LOCATION)
			IoStackLocation)->Parameters.QueryDirectory.FileName;
		
		FileIndex = ((PEXTENDED_IO_STACK_LOCATION)
			IoStackLocation)->Parameters.QueryDirectory.FileIndex;
		
#endif // _GNU_NTIFS_
		
		RestartScan = IoStackLocation->Flags & SL_RESTART_SCAN;
		ReturnSingleEntry = IoStackLocation->Flags & SL_RETURN_SINGLE_ENTRY;
		IndexSpecified = IoStackLocation->Flags & SL_INDEX_SPECIFIED;
		
		if (!IrpContext->IsSynchronous)
		{
			Status = STATUS_PENDING;
			__leave;
		}
		
		Buffer = Ext2GetUserBuffer(Irp);
		
		if (Buffer == NULL)
		{
			Status = STATUS_INVALID_USER_BUFFER;
			__leave;
		}
		
		if (FileName != NULL)
		{
	
			if (Ccb->DirectorySearchPattern.Buffer != NULL)
			{
				FirstQuery = FALSE;
			}
			else
			{
				FirstQuery = TRUE;
				
				Ccb->DirectorySearchPattern.Length =
					Ccb->DirectorySearchPattern.MaximumLength =
					FileName->Length;
				
				Ccb->DirectorySearchPattern.Buffer =
					ExAllocatePool(NonPagedPool, FileName->Length);
				
				if (Ccb->DirectorySearchPattern.Buffer == NULL)
				{
					Status = STATUS_INSUFFICIENT_RESOURCES;
					__leave;
				}
				
				RtlCopyMemory(
					Ccb->DirectorySearchPattern.Buffer,
					FileName->Buffer,
					FileName->Length );
			}
		}
		else if (Ccb->DirectorySearchPattern.Buffer != NULL)
		{
			FirstQuery = FALSE;
			FileName = &Ccb->DirectorySearchPattern;
		}
		else
		{
			FirstQuery = TRUE;
			
			Ccb->DirectorySearchPattern.Length =
				Ccb->DirectorySearchPattern.MaximumLength =
				sizeof(L"*");
			
			Ccb->DirectorySearchPattern.Buffer =
				ExAllocatePool(NonPagedPool, sizeof(L"*"));
			
			if (Ccb->DirectorySearchPattern.Buffer == NULL)
			{
				Status = STATUS_INSUFFICIENT_RESOURCES;
				__leave;
			}
			
			RtlCopyMemory(
				Ccb->DirectorySearchPattern.Buffer,
				L"*", sizeof(L"*"));
		}
		
		if (!IndexSpecified)
		{
			if (RestartScan || FirstQuery)
			{
				FileIndex = (ULONG)Fcb->IndexNumber.QuadPart;
			}
			else
			{
				FileIndex = Ccb->CurrentByteOffset;
			}
		}
		
		Inode = (struct ext2_inode*) ExAllocatePool(
			NonPagedPool,
			sizeof(struct ext2_inode));
		
		if (Inode == NULL)
		{
			Status = STATUS_INSUFFICIENT_RESOURCES;
			__leave;
		}
		
		RtlZeroMemory(Buffer, Length);

		dwSize = Fcb->ext2_inode->i_size - FileIndex;
		if (dwSize <= 0)
		{
			Status = STATUS_NO_MORE_FILES;
			__leave;
		}
		
		DirectoryContent = ExAllocatePool(NonPagedPool,
			dwSize);
		if (!DirectoryContent)
		{
			Status = STATUS_INSUFFICIENT_RESOURCES;
			__leave;
		}
		
		dwBytes = 0;
		RtlZeroMemory(DirectoryContent, dwSize);
		Ext2ReadInode(Vcb, Fcb->ext2_inode, FileIndex, DirectoryContent, dwSize);
		dwTemp = 0;
		
		pDir = (struct ext2_dir_entry *) ((PUCHAR)DirectoryContent + dwBytes);

		while (bRun && UsedLength < Length  && dwBytes < dwSize && pDir->inode)
		{

			InodeFileNameLength = pDir->name_len & 0xff;
			
			InodeFileName.Length = InodeFileName.MaximumLength =
				InodeFileNameLength * 2;
			
			if (InodeFileName.Length <=0 )
				break;

			InodeFileName.Buffer = ExAllocatePool(
				NonPagedPool,
				InodeFileNameLength * 2 + 2);

			RtlZeroMemory(InodeFileName.Buffer, InodeFileNameLength * 2 + 2);
			
			Ext2CharToWchar(
				InodeFileName.Buffer,
				pDir->name,
				InodeFileNameLength );

//			KdPrint(("Ext2QueryDirectory: Found %S to comp.\n", InodeFileName.Buffer));
/*			
			if (FsRtlIsNameInExpression(
				FileName,
				&InodeFileName,
				FALSE,
				NULL ))
*/
			if (FsRtlDoesNameContainWildCards(FileName) ?
				FsRtlIsNameInExpression(
					FileName,
					&InodeFileName,
					TRUE,
					NULL) :
				!RtlCompareUnicodeString(
					FileName,
					&InodeFileName,
					TRUE) 			)
			{
				dwReturn = Ext2ProcessDirEntry(
					Vcb, FileInformationClass,
					pDir->inode, Buffer, UsedLength, Length - UsedLength,
					dwBytes, &InodeFileName,
					ReturnSingleEntry );
				if (dwReturn <= 0)
				{
					bRun = FALSE;
				}
				else
				{
					dwTemp = UsedLength;
					UsedLength += dwReturn;
				}
			}
			
			if (InodeFileName.Buffer != NULL)
			{
				ExFreePool(InodeFileName.Buffer);
				InodeFileName.Buffer = NULL;
			}
			
			if (bRun)
			{
				dwBytes +=pDir->rec_len;
				Ccb->CurrentByteOffset = FileIndex + dwBytes;
				pDir = (struct ext2_dir_entry *) ((PUCHAR)DirectoryContent + dwBytes);
			}

			if (UsedLength && ReturnSingleEntry)
			{
				Status = STATUS_SUCCESS;
				__leave;
			}
		}

		FileIndex += dwBytes;

		((PULONG)((PUCHAR)Buffer + dwTemp)) [0] = 0;

		if (!UsedLength)
		{
			if (FirstQuery)
			{
				Status = STATUS_NO_SUCH_FILE;
			}
			else
			{
				Status = STATUS_NO_MORE_FILES;
			}
		}
		else
		{
			Status = STATUS_SUCCESS;
		}
	}

	__finally
	{
	
		if (FcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread()
				);
		}
		
		if (Inode != NULL)
		{
			ExFreePool(Inode);
		}
		
		if (DirectoryContent != NULL)
		{
			ExFreePool(DirectoryContent);
			DirectoryContent = NULL;
		}
		
		if (InodeFileName.Buffer != NULL)
		{
			ExFreePool(InodeFileName.Buffer);
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			if (Status == STATUS_PENDING)
			{
				Status = Ext2LockUserBuffer(
					IrpContext->Irp,
					Length,
					IoWriteAccess );
				
				if (NT_SUCCESS(Status))
				{
					Status = Ext2QueueRequest(IrpContext);
				}
				else
				{
					IrpContext->Irp->IoStatus.Status = Status;
					Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
					Ext2FreeIrpContext(IrpContext);
				}
			}
			else
			{
				IrpContext->Irp->IoStatus.Information = UsedLength;
				IrpContext->Irp->IoStatus.Status = Status;
				
				Ext2CompleteRequest(
					IrpContext->Irp,
					(CCHAR)
					(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
				
				Ext2FreeIrpContext(IrpContext);
			}
		}
	}
	
	return Status;
}


NTSTATUS
Ext2DirectoryControl (IN PEXT2_IRP_CONTEXT IrpContext)
{
	NTSTATUS Status;
	
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	switch (IrpContext->MinorFunction)
	{
	case IRP_MN_QUERY_DIRECTORY:
		Status = Ext2QueryDirectory(IrpContext);
		break;
		
	default:
		Status = STATUS_INVALID_DEVICE_REQUEST;
		IrpContext->Irp->IoStatus.Status = Status;
		Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
		Ext2FreeIrpContext(IrpContext);
	}
	
	return Status;
}


NTSTATUS
Ext2QueueRequest (IN PEXT2_IRP_CONTEXT IrpContext)
{
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	// IsSynchronous means we can block (so we don't requeue it)
	IrpContext->IsSynchronous = TRUE;
	
	IoMarkIrpPending(IrpContext->Irp);
	
	ExInitializeWorkItem(
		&IrpContext->WorkQueueItem,
		Ext2DeQueueRequest,
		IrpContext );
	
	ExQueueWorkItem(&IrpContext->WorkQueueItem, CriticalWorkQueue);
	
	return STATUS_PENDING;
}


VOID
Ext2DeQueueRequest (IN PVOID Context)
{
    PEXT2_IRP_CONTEXT IrpContext;

    IrpContext = (PEXT2_IRP_CONTEXT) Context;

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));

    __try
    {
        __try
        {
            FsRtlEnterFileSystem();

            if (!IrpContext->IsTopLevel)
            {
                IoSetTopLevelIrp((PIRP) FSRTL_FSP_TOP_LEVEL_IRP);
            }

            Ext2DispatchRequest(IrpContext);
        }
        __except (Ext2ExceptionFilter(IrpContext, GetExceptionCode()))
        {
            Ext2ExceptionHandler(IrpContext);
        }
    }
    __finally
    {
        IoSetTopLevelIrp(NULL);

        FsRtlExitFileSystem();
    }
}


NTSTATUS
Ext2Cleanup (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT  DeviceObject;
	NTSTATUS        Status = STATUS_SUCCESS;
	PEXT2_VCB       Vcb;
	BOOLEAN         VcbResourceAcquired = FALSE;
	PFILE_OBJECT    FileObject;
	PEXT2_FCB       Fcb;
	BOOLEAN         FcbResourceAcquired = FALSE;
	PEXT2_CCB       Ccb;
	PIRP            Irp;

	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_SUCCESS;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		FileObject = IrpContext->FileObject;
		
		Fcb = (PEXT2_FCB) FileObject->FsContext;
		
		ASSERT(Fcb != NULL);
		
		if (Fcb->Identifier.Type == VCB)
		{
			Status = STATUS_SUCCESS;
			__leave;
		}
		
		ASSERT((Fcb->Identifier.Type == FCB) &&
			(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
		
		Ccb = (PEXT2_CCB) FileObject->FsContext2;
		
		ASSERT(Ccb != NULL);
		
		ASSERT((Ccb->Identifier.Type == CCB) &&
			(Ccb->Identifier.Size == sizeof(EXT2_CCB)));
		
		Irp = IrpContext->Irp;

		if (!ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			IrpContext->IsSynchronous
			))
		{
			Status = STATUS_PENDING;
			__leave;
		}
		
		VcbResourceAcquired = TRUE;
		
#ifndef EXT2_RO
		if (!Fcb->IsPageFile)
#endif
		{
			if (!ExAcquireResourceExclusiveLite(
				&Fcb->MainResource,
				IrpContext->IsSynchronous
				))
			{
				Status = STATUS_PENDING;
				__leave;
			}
			
			FcbResourceAcquired = TRUE;
		}
		
		Fcb->OpenHandleCount--;
		
		Vcb->OpenFileHandleCount--;
		
		CcUninitializeCacheMap(FileObject, NULL, NULL);
		
		if (!(Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY))
		{
			//
			// Drop any byte range locks this process may have on the file.
			//
			FsRtlFastUnlockAll(
				&Fcb->FileLockAnchor,
				FileObject,
				IoGetRequestorProcess(Irp),
				NULL );
			
			//
			// If there are no byte range locks owned by other processes on the
			// file the fast I/O read/write functions doesn't have to check for
			// locks so we set IsFastIoPossible to FastIoIsPossible again.
			//
			if (!FsRtlGetNextFileLock(&Fcb->FileLockAnchor, TRUE))
			{
				if (Fcb->CommonFCBHeader.IsFastIoPossible != FastIoIsPossible)
				{
#if DBG
					KdPrint(("%-16.16s %-31s %s\n",
						PsGetCurrentProcess()->ImageFileName,
						"FastIoIsPossible",
						Fcb->DbgFileName));
#endif					
					Fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsPossible;
				}
			}
		}
		
#ifndef EXT2_RO
		
		IoRemoveShareAccess(FileObject, &Fcb->ShareAccess);
		
#endif
#if DBG		
		KdPrint(("Ext2Cleanup: OpenCount: %u ReferCount: %-7u %s\n",
			Fcb->OpenHandleCount,
			Fcb->ReferenceCount,
			Fcb->DbgFileName ));
#endif
		Status = STATUS_SUCCESS;
	}
	__finally
	{
		if (IrpContext->FileObject)
		{
			IrpContext->FileObject->Flags |= FO_CLEANUP_COMPLETE;
		}
		
		if (FcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread() );
		}
		
		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread());
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			if (Status == STATUS_PENDING)
			{
				Ext2QueueRequest(IrpContext);
			}
			else
			{
				IrpContext->Irp->IoStatus.Status = Status;
				
				Ext2CompleteRequest(
					IrpContext->Irp,
					(CCHAR)
					(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
				
				Ext2FreeIrpContext(IrpContext);
			}
		}
	}
	
	return Status;
}

#if DBG

NTSTATUS
Ext2DeviceControlCompletion (IN PDEVICE_OBJECT   DeviceObject,
			     IN PIRP             Irp,
			     IN PVOID            Context)
{
	if (Irp->PendingReturned)
	{
		IoMarkIrpPending(Irp);
	}
	
	KdPrint((DRIVER_NAME ": %-16.16s %-31s *** Status: %s (%#x) ***\n",
		PsGetCurrentProcess()->ImageFileName,
		"IRP_MJ_DEVICE_CONTROL",
		Ext2NtStatusToString(Irp->IoStatus.Status),
		Irp->IoStatus.Status ));
	
	return STATUS_SUCCESS;
}

#endif // DBG


NTSTATUS
Ext2DeviceControlNormal (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT  DeviceObject;
	BOOLEAN         CompleteRequest;
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB       Vcb;
	PIRP            Irp;
	PDEVICE_OBJECT  TargetDeviceObject;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
	
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			CompleteRequest = TRUE;
			Status = STATUS_INVALID_DEVICE_REQUEST;
			
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		Irp = IrpContext->Irp;
		
		TargetDeviceObject = Vcb->TargetDeviceObject;
		
		//
		// Pass on the IOCTL to the driver below
		//
		
		CompleteRequest = FALSE;

		IoSkipCurrentIrpStackLocation(Irp);
#if DBG
		IoSetCompletionRoutine(
			Irp,
			Ext2DeviceControlCompletion,
			DeviceObject,
			FALSE,
			TRUE,
			TRUE );
#endif
		
		Status = IoCallDriver(TargetDeviceObject, Irp);
	}
	__finally
	{
		if (!IrpContext->ExceptionInProgress)
		{
			if (CompleteRequest)
			{
				IrpContext->Irp->IoStatus.Status = Status;
				
				Ext2CompleteRequest(
					IrpContext->Irp,
					(CCHAR)
					(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
					);
			}

			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}


NTSTATUS
Ext2PrepareToUnload (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT  DeviceObject;
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	BOOLEAN         GlobalDataResourceAcquired = FALSE;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		
		if (DeviceObject != Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		ExAcquireResourceExclusiveLite(
			&Ext2Global->Resource,
			TRUE );
		
		GlobalDataResourceAcquired = TRUE;
		
		if (Ext2Global->Flags & EXT2_UNLOAD_PENDING)
		{
			KdPrint(("Ext2PrepareUnload:  Already ready to unload.\n"));
			
			Status = STATUS_ACCESS_DENIED;
			
			__leave;
		}
		
		if (!IsListEmpty(&(Ext2Global->VcbList)))
		{
			KdPrint(("Ext2PrepareUnload:  Mounted volumes exists.\n"));
			
			Status = STATUS_ACCESS_DENIED;
			
			__leave;
		}
		
		IoUnregisterFileSystem(Ext2Global->DeviceObject);
		
		Ext2Global->DriverObject->DriverUnload = DriverUnload;
		
		Ext2Global->Flags |= EXT2_UNLOAD_PENDING;
		
//		KdPrint(("Ext2PrepareToUnload: Driver is ready to unload.\n"));
		
		Status = STATUS_SUCCESS;
	}
	__finally
	{
		if (GlobalDataResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Ext2Global->Resource,
				ExGetCurrentResourceThread()
				);
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(
				IrpContext->Irp,
				(CCHAR)
				(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
			
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}



NTSTATUS
Ext2DeviceControl (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PIRP                Irp;
	PIO_STACK_LOCATION  IoStackLocation;
	ULONG               IoControlCode;
	NTSTATUS            Status;
	
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	Irp = IrpContext->Irp;
	
	IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
	
	IoControlCode =
		IoStackLocation->Parameters.DeviceIoControl.IoControlCode;
	
	switch (IoControlCode)
	{
	case IOCTL_PREPARE_TO_UNLOAD:
		Status = Ext2PrepareToUnload(IrpContext);
		break;
		
	default:
		Status = Ext2DeviceControlNormal(IrpContext);
	}
	
	return Status;
}


NTSTATUS
Ext2VerifyVolume (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT          DeviceObject;
	NTSTATUS                Status = STATUS_UNSUCCESSFUL;
	PEXT2_SUPER_BLOCK	ext2_sb = NULL;
	PEXT2_VCB               Vcb;
	BOOLEAN                 VcbResourceAcquired = FALSE;
	PIRP                    Irp;
	PIO_STACK_LOCATION      IoStackLocation;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			TRUE );
		
		VcbResourceAcquired = TRUE;
		
		if (!(Vcb->TargetDeviceObject->Flags & DO_VERIFY_VOLUME))
		{
			Status = STATUS_SUCCESS;
			__leave;
		}
		
		Irp = IrpContext->Irp;
		
		IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
		
		ext2_sb = Ext2LoadSuper(Vcb->TargetDeviceObject);

		if (!ext2_sb)
		{
			__leave;
		}

		if (ext2_sb->s_magic != EXT2_SUPER_MAGIC)
		{
			__leave;
		}

		Vcb->TargetDeviceObject->Flags &= ~DO_VERIFY_VOLUME;
			
//		KdPrint(("Ext2VerifyVolume: Volume verify succeeded.\n"));

		Status = STATUS_SUCCESS;
			
		__leave;
	}
	__finally
	{
		if (ext2_sb)
			ExFreePool(ext2_sb);

		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread()
				);
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(
				IrpContext->Irp,
				(CCHAR)
				(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
			
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}


NTSTATUS
Ext2IsVolumeMounted (IN PEXT2_IRP_CONTEXT IrpContext)
{
    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));

    return Ext2VerifyVolume(IrpContext);
}


NTSTATUS
Ext2DismountVolume (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT  DeviceObject;
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB       Vcb;
	BOOLEAN         VcbResourceAcquired = FALSE;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;

		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			TRUE );
		
		VcbResourceAcquired = TRUE;
		
		if (!(Vcb->Flags & VCB_VOLUME_LOCKED))
		{
			KdPrint(("Ext2Dismount: Volume is not locked.\n"));
			
			Status = STATUS_ACCESS_DENIED;
			
			__leave;
		}
		
		Vcb->Flags |= VCB_DISMOUNT_PENDING;
		
		KdPrint(("Ext2Dismount: Volume dismount pending.\n"));
		
		Status = STATUS_SUCCESS;
	}
	__finally
	{
		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread()
				);
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(
				IrpContext->Irp,
				(CCHAR)
				(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
				);
			
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}


VOID
Ext2PurgeVolume (IN PEXT2_VCB Vcb,
		 IN BOOLEAN  FlushBeforePurge )
{
	BOOLEAN         VcbResourceAcquired = FALSE;
	PEXT2_FCB       Fcb;
	LIST_ENTRY      FcbList;
	PLIST_ENTRY     ListEntry;
	PFCB_LIST_ENTRY FcbListEntry;
	
	__try
	{
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		ExAcquireResourceSharedLite(
			&Vcb->MainResource,
			TRUE );
		
		VcbResourceAcquired = TRUE;
		
#ifndef EXT2_RO
		if (Vcb->ReadOnly)
		{
			Flush = FALSE;
		}
#endif
		
		InitializeListHead(&FcbList);
		
		for (
			ListEntry = Vcb->FcbList.Flink;
		ListEntry != &Vcb->FcbList;
		ListEntry = ListEntry->Flink
			)
		{
			Fcb = CONTAINING_RECORD(ListEntry, EXT2_FCB, Next);
			
			ExAcquireResourceExclusiveLite(
				&Fcb->MainResource,
				TRUE );
			
			Fcb->ReferenceCount++;
#if DBG
			KdPrint(("Ext2PurgeVolume: %s refercount=%xh\n", Fcb->DbgFileName, Fcb->ReferenceCount));
#endif	
			
			ExReleaseResourceForThreadLite(
				&Fcb->MainResource,
				ExGetCurrentResourceThread());
			
			FcbListEntry = ExAllocatePool(NonPagedPool, sizeof(FCB_LIST_ENTRY));
			
			FcbListEntry->Fcb = Fcb;
			
			InsertTailList(&FcbList, &FcbListEntry->Next);
		}
		
		ExReleaseResourceForThreadLite(
			&Vcb->MainResource,
			ExGetCurrentResourceThread() );
		
		VcbResourceAcquired = FALSE;
		
		while (!IsListEmpty(&FcbList))
		{
			ListEntry = RemoveHeadList(&FcbList);
			
			FcbListEntry = CONTAINING_RECORD(ListEntry, FCB_LIST_ENTRY, Next);
			
			Fcb = FcbListEntry->Fcb;
			
			Ext2PurgeFile(Fcb, FlushBeforePurge);
			
			if (!Fcb->OpenHandleCount && Fcb->ReferenceCount == 1)
			{
#if DBG
				KdPrint(("Ext2FreeFcb %s.\n", Fcb->DbgFileName));
#endif
				Ext2FreeFcb(Fcb);
			}
			
			ExFreePool(FcbListEntry);
		}
		
		KdPrint(("Ext2PurgeVolume: Volume flushed and purged.\n"));
	}
	__finally
	{
		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread() );
		}
	}
}

VOID
Ext2PurgeFile (IN PEXT2_FCB Fcb,
	       IN BOOLEAN  FlushBeforePurge )
{
	ASSERT(Fcb != NULL);
        
	ASSERT((Fcb->Identifier.Type == FCB) &&
		(Fcb->Identifier.Size == sizeof(EXT2_FCB)));
	
#ifndef EXT2_RO
	if (FlushBeforePurge)
	{
#if DBG
		KdPrint(("Ext2PurgeFile: CcFlushCache on %s.\n", Fcb->DbgFileName));
#endif
		
		CcFlushCache(&Fcb->SectionObject, NULL, 0, &IoStatus);
	}
#endif
	
	if (Fcb->SectionObject.ImageSectionObject)
	{
#if DBG
		KdPrint(("Ext2PurgeFile: MmFlushImageSection on %s.\n", Fcb->DbgFileName));
#endif
	
		MmFlushImageSection(&Fcb->SectionObject, MmFlushForWrite);
	}
	
	if (Fcb->SectionObject.DataSectionObject)
	{
#if DBG
		KdPrint(("Ext2PurgeFile: CcPurgeCacheSection on %s.\n", Fcb->DbgFileName));
#endif		
		CcPurgeCacheSection(&Fcb->SectionObject, NULL, 0, FALSE);
	}
}



NTSTATUS
Ext2LockVolume (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT  DeviceObject;
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB       Vcb;
	BOOLEAN         VcbResourceAcquired = FALSE;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		
		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		ExAcquireResourceSharedLite(
			&Vcb->MainResource,
			TRUE
			);
		
		VcbResourceAcquired = TRUE;
		
		if (Vcb->Flags & VCB_VOLUME_LOCKED)
		{
			KdPrint(("Ext2LockVolume: Volume is already locked.\n"));
			
			Status = STATUS_ACCESS_DENIED;
			
			__leave;
		}
		
		if (Vcb->OpenFileHandleCount)
		{
			KdPrint(("Ext2LockVolume: Open files exists.\n"));
			
			Status = STATUS_ACCESS_DENIED;
			
			__leave;
		}
		
		ExReleaseResourceForThreadLite(
			&Vcb->MainResource,
			ExGetCurrentResourceThread()
			);
		
		VcbResourceAcquired = FALSE;
		
		Ext2PurgeVolume(Vcb, TRUE);
		
		ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			TRUE
			);
		
		VcbResourceAcquired = TRUE;
		
		if (!IsListEmpty(&Vcb->FcbList))
		{
			KdPrint(("Ext2LockVolume: Could not purge cached files.\n"));
			
			Status = STATUS_ACCESS_DENIED;
			
			__leave;
		}
		
		Vcb->Flags |= VCB_VOLUME_LOCKED;
		
		Ext2SetVpbFlag(Vcb->Vpb, VPB_LOCKED);
		
		KdPrint(("Ext2LockVolume: Volume locked.\n"));
		
		Status = STATUS_SUCCESS;
	}
	__finally
	{
		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread());
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(
				IrpContext->Irp,
				(CCHAR)
				(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
			
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}

NTSTATUS
Ext2UnlockVolume (
		 IN PEXT2_IRP_CONTEXT IrpContext
		 )
{
	PDEVICE_OBJECT  DeviceObject;
	NTSTATUS        Status = STATUS_UNSUCCESSFUL;
	PEXT2_VCB       Vcb;
	BOOLEAN         VcbResourceAcquired = FALSE;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		DeviceObject = IrpContext->DeviceObject;
		
		//
		// This request is not allowed on the main device object
		//
		if (DeviceObject == Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
		
		ASSERT(Vcb != NULL);
		
		ASSERT((Vcb->Identifier.Type == VCB) &&
			(Vcb->Identifier.Size == sizeof(EXT2_VCB)));
		
		ExAcquireResourceExclusiveLite(
			&Vcb->MainResource,
			TRUE );
		
		VcbResourceAcquired = TRUE;
		
		if (!(Vcb->Flags & VCB_VOLUME_LOCKED))
		{
			KdPrint((": Ext2UnlockVolume: Volume is not locked .\n"));
			
			Status = STATUS_ACCESS_DENIED;
			
			__leave;
		}
		
		Vcb->Flags &= ~VCB_VOLUME_LOCKED;
		
		Ext2ClearVpbFlag(Vcb->Vpb, VPB_LOCKED);
		
		KdPrint(("Ext2UnlockVolume: Volume unlocked.\n"));
		
		Status = STATUS_SUCCESS;
	}
	__finally
	{
		if (VcbResourceAcquired)
		{
			ExReleaseResourceForThreadLite(
				&Vcb->MainResource,
				ExGetCurrentResourceThread()
				);
		}
		
		if (!IrpContext->ExceptionInProgress)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(
				IrpContext->Irp,
				(CCHAR)
				(NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
			
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}


NTSTATUS
Ext2UserFsRequest (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PIRP                Irp;
	PIO_STACK_LOCATION  IoStackLocation;
	ULONG               FsControlCode;
	NTSTATUS            Status;
	
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	Irp = IrpContext->Irp;
	
	IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
	
#ifndef _GNU_NTIFS_
	FsControlCode =
		IoStackLocation->Parameters.FileSystemControl.FsControlCode;
#else
	FsControlCode = ((PEXTENDED_IO_STACK_LOCATION)
		IoStackLocation)->Parameters.FileSystemControl.FsControlCode;
#endif
	
	switch (FsControlCode)
	{
	case FSCTL_LOCK_VOLUME:
		Status = Ext2LockVolume(IrpContext);
		break;
		
	case FSCTL_UNLOCK_VOLUME:
		Status = Ext2UnlockVolume(IrpContext);
		break;
		
	case FSCTL_DISMOUNT_VOLUME:
		Status = Ext2DismountVolume(IrpContext);
		break;
		
	case FSCTL_IS_VOLUME_MOUNTED:
		Status = Ext2IsVolumeMounted(IrpContext);
		break;
		
	default:
		Status = STATUS_INVALID_DEVICE_REQUEST;
		IrpContext->Irp->IoStatus.Status = Status;
		Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
		Ext2FreeIrpContext(IrpContext);
	}
	
	return Status;
}

NTSTATUS
Ext2MountVolume (IN PEXT2_IRP_CONTEXT IrpContext)
{
	PDEVICE_OBJECT              MainDeviceObject;
	BOOLEAN                     GlobalDataResourceAcquired = FALSE;
	PIRP                        Irp;
	PIO_STACK_LOCATION          IoStackLocation;
	PDEVICE_OBJECT              TargetDeviceObject;
	NTSTATUS                    Status = STATUS_UNSUCCESSFUL;
	PDEVICE_OBJECT              VolumeDeviceObject = NULL;
	PEXT2_VCB                   Vcb;
	BOOLEAN                     VcbResourceInitialized = FALSE;
	struct ext2_super_block*    ext2_super_block = NULL;
	struct ext2_group_desc *    ext2_group_desc = NULL;
	USHORT                      VolumeLabelLength;
	ULONG                       IoctlSize;
	
	__try
	{
		ASSERT(IrpContext != NULL);
		
		ASSERT((IrpContext->Identifier.Type == ICX) &&
			(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
		
		MainDeviceObject = IrpContext->DeviceObject;

		//
		// This request is only allowed on the main device object
		//
		if (MainDeviceObject != Ext2Global->DeviceObject)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			__leave;
		}
		
		ExAcquireResourceExclusiveLite(
			&(Ext2Global->Resource),
			TRUE );
		
		GlobalDataResourceAcquired = TRUE;
		
		if (Ext2Global->Flags & EXT2_UNLOAD_PENDING)
		{
			Status = STATUS_UNRECOGNIZED_VOLUME;
			__leave;
		}
		
		Irp = IrpContext->Irp;
		
		IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
		
		TargetDeviceObject =
			IoStackLocation->Parameters.MountVolume.DeviceObject;
		
		ext2_super_block = Ext2LoadSuper(TargetDeviceObject);

		Status = STATUS_UNSUCCESSFUL;
		if (ext2_super_block)
		{
			if (ext2_super_block->s_magic == EXT2_SUPER_MAGIC)
			{
				KdPrint(("Ext2Fsd: Ext2fs is found.\n"));
				Status = STATUS_SUCCESS;
			}
		}

		if (!NT_SUCCESS(Status))
		{
			__leave;
		}
		
		Status = IoCreateDevice(
			MainDeviceObject->DriverObject,
			sizeof(EXT2_VCB),
			NULL,
			FILE_DEVICE_DISK_FILE_SYSTEM,
			0,
			FALSE,
			&VolumeDeviceObject );
		
		if (!NT_SUCCESS(Status))
		{
			__leave;
		}
		
		VolumeDeviceObject->StackSize = TargetDeviceObject->StackSize;
		
		(IoStackLocation->Parameters.MountVolume.Vpb)->DeviceObject =
			VolumeDeviceObject;
		
		Vcb = (PEXT2_VCB) VolumeDeviceObject->DeviceExtension;
		
		RtlZeroMemory(Vcb, sizeof(EXT2_VCB));
		
		Vcb->Identifier.Type = VCB;
		Vcb->Identifier.Size = sizeof(EXT2_VCB);

#ifndef EXT2_RO
		Vcb->ReadOnly = TRUE;
#endif
		
		ExInitializeResourceLite(&Vcb->MainResource);
		ExInitializeResourceLite(&Vcb->PagingIoResource);
		
		VcbResourceInitialized = TRUE;
		
		Vcb->Vpb = IoStackLocation->Parameters.MountVolume.Vpb;
		
		InitializeListHead(&Vcb->FcbList);
		
		Vcb->DeviceObject = VolumeDeviceObject;
		
		Vcb->TargetDeviceObject = TargetDeviceObject;
		
		Vcb->OpenFileHandleCount = 0;
		
		Vcb->ReferenceCount = 0;
		
		Vcb->Flags = 0;
		
		Vcb->ext2_super_block = ext2_super_block;
		
		VolumeLabelLength = 0;
		
		if (VolumeLabelLength > MAXIMUM_VOLUME_LABEL_LENGTH / 2)
		{
			VolumeLabelLength = MAXIMUM_VOLUME_LABEL_LENGTH / 2;
		}
		
		Vcb->Vpb->VolumeLabelLength = VolumeLabelLength * 2;
/*		
		Ext2CharToWchar(
			Vcb->Vpb->VolumeLabel,
			Vcb->romfs_super_block->name,
			VolumeLabelLength
			);
*/
		
		Vcb->Vpb->SerialNumber = 'MATT';


		IoctlSize = sizeof(DISK_GEOMETRY);
		
		Status = Ext2DiskIoControl(
			TargetDeviceObject,
			IOCTL_DISK_GET_DRIVE_GEOMETRY,
			NULL,
			0,
			&Vcb->DiskGeometry,
			&IoctlSize );
		
		if (!NT_SUCCESS(Status))
		{
			__leave;
		}
		
		IoctlSize = sizeof(PARTITION_INFORMATION);
		
		Status = Ext2DiskIoControl(
			TargetDeviceObject,
			IOCTL_DISK_GET_PARTITION_INFO,
			NULL,
			0,
			&Vcb->PartitionInformation,
			&IoctlSize );
		
		if (!NT_SUCCESS(Status))
		{
			Vcb->PartitionInformation.StartingOffset.QuadPart = 0;
			
			Vcb->PartitionInformation.PartitionLength.QuadPart =
				Vcb->DiskGeometry.Cylinders.QuadPart *
				Vcb->DiskGeometry.TracksPerCylinder *
				Vcb->DiskGeometry.SectorsPerTrack *
				Vcb->DiskGeometry.BytesPerSector;
			
			Status = STATUS_SUCCESS;
		}
		
		ext2_group_desc = Ext2LoadGroup(Vcb); 
		if (!ext2_group_desc)
		{
			Status = STATUS_UNSUCCESSFUL;
			__leave;
		}

		Vcb->ext2_group_desc = ext2_group_desc;
		InsertTailList(&(Ext2Global->VcbList), &Vcb->Next);
    }

    __finally
    {
	    if (GlobalDataResourceAcquired)
	    {
		    ExReleaseResourceForThreadLite(
			    &Ext2Global->Resource,
			    ExGetCurrentResourceThread() );
	    }
	    
	    if (!NT_SUCCESS(Status))
	    {
		    if (ext2_super_block)
		    {
			    ExFreePool(ext2_super_block);
		    }
		    
		    if (VcbResourceInitialized)
		    {
			    ExDeleteResourceLite(&Vcb->MainResource);
			    ExDeleteResourceLite(&Vcb->PagingIoResource);
		    }
		    
		    if (VolumeDeviceObject)
		    {
			    IoDeleteDevice(VolumeDeviceObject);
		    }
	    }
	    
	    if (!IrpContext->ExceptionInProgress)
	    {
		    if (NT_SUCCESS(Status))
		    {
			    VolumeDeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;
		    }
		    
		    IrpContext->Irp->IoStatus.Status = Status;
		    
		    Ext2CompleteRequest(
			    IrpContext->Irp,
			    (CCHAR)
			    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
		    
		    Ext2FreeIrpContext(IrpContext);
	    }
    }
    
    return Status;
}


NTSTATUS
Ext2FileSystemControl (IN PEXT2_IRP_CONTEXT IrpContext)
{
	NTSTATUS Status;
	
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	switch (IrpContext->MinorFunction)
	{
	case IRP_MN_USER_FS_REQUEST:
		Status = Ext2UserFsRequest(IrpContext);
		break;
		
	case IRP_MN_MOUNT_VOLUME:
		Status = Ext2MountVolume(IrpContext);
		break;
		
	case IRP_MN_VERIFY_VOLUME:
		Status = Ext2VerifyVolume(IrpContext);
		break;
		
	default:
		Status = STATUS_INVALID_DEVICE_REQUEST;
		IrpContext->Irp->IoStatus.Status = Status;
		Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
		Ext2FreeIrpContext(IrpContext);
	}
	
	return Status;
}


NTSTATUS
Ext2DispatchRequest (IN PEXT2_IRP_CONTEXT IrpContext)
{
	ASSERT(IrpContext);
	
	ASSERT((IrpContext->Identifier.Type == ICX) &&
		(IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
	
	switch (IrpContext->MajorFunction)
	{
	case IRP_MJ_CREATE:
		return Ext2Create(IrpContext);
		
	case IRP_MJ_CLOSE:
		return Ext2Close(IrpContext);
		
	case IRP_MJ_READ:
		return Ext2Read(IrpContext);
		
	case IRP_MJ_QUERY_INFORMATION:
		return Ext2QueryInformation(IrpContext);
		
	case IRP_MJ_SET_INFORMATION:
		return Ext2SetInformation(IrpContext);
		
	case IRP_MJ_QUERY_VOLUME_INFORMATION:
		return Ext2QueryVolumeInformation(IrpContext);
		
	case IRP_MJ_DIRECTORY_CONTROL:
		return Ext2DirectoryControl(IrpContext);
		
	case IRP_MJ_FILE_SYSTEM_CONTROL:
		return Ext2FileSystemControl(IrpContext);
		
	case IRP_MJ_DEVICE_CONTROL:
		return Ext2DeviceControl(IrpContext);

	case IRP_MJ_LOCK_CONTROL:
		return Ext2LockControl(IrpContext);
		
	case IRP_MJ_CLEANUP:
		return Ext2Cleanup(IrpContext);
		
	default:
		KdPrint(("Ext2DispatchRequest: Unexpected major function: %#x\n",
			IrpContext->MajorFunction));
		IrpContext->Irp->IoStatus.Status = STATUS_DRIVER_INTERNAL_ERROR;
		Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
		Ext2FreeIrpContext(IrpContext);
		return STATUS_DRIVER_INTERNAL_ERROR;
	}
}

NTSTATUS
Ext2ExceptionFilter (IN PEXT2_IRP_CONTEXT    IrpContext,
		     IN NTSTATUS             ExceptionCode)
{
	NTSTATUS Status;
	
	//
	// Only use a valid IrpContext
	//
	if (IrpContext)
	{
		if ((IrpContext->Identifier.Type != ICX) ||
			(IrpContext->Identifier.Size != sizeof(EXT2_IRP_CONTEXT)))
		{
			IrpContext = NULL;
		}
	}
	
	//
	// If the exception is expected execute our handler
	//
	if (FsRtlIsNtstatusExpected(ExceptionCode))
	{
		KdPrint(("Ext2ExceptionFilter: Catching exception %#x\n",
			ExceptionCode));
		
		Status = EXCEPTION_EXECUTE_HANDLER;
		
		if (IrpContext)
		{
			IrpContext->ExceptionInProgress = TRUE;
			IrpContext->ExceptionCode = ExceptionCode;
		}
	}
	//
	// else continue search for an higher level exception handler
	//
	else
	{
		KdPrint(("Ext2ExceptionFilter: Passing on exception %#x\n",
			ExceptionCode));
		
		Status = EXCEPTION_CONTINUE_SEARCH;
		
		if (IrpContext)
		{
			Ext2FreeIrpContext(IrpContext);
		}
	}
	
	return Status;
}

NTSTATUS
Ext2ExceptionHandler (IN PEXT2_IRP_CONTEXT IrpContext)
{
	NTSTATUS Status;
	
	if (IrpContext)
	{
		Status = IrpContext->ExceptionCode;
		
		if (IrpContext->Irp)
		{
			IrpContext->Irp->IoStatus.Status = Status;
			
			Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
		}
		
		Ext2FreeIrpContext(IrpContext);
	}
	else
	{
		Status = STATUS_INSUFFICIENT_RESOURCES;
	}
	
	return Status;
}


VOID
Ext2SetVpbFlag (IN PVPB     Vpb,
		IN USHORT   Flag )
{
	KIRQL OldIrql;
	
	IoAcquireVpbSpinLock(&OldIrql);
	
	Vpb->Flags |= Flag;
	
	IoReleaseVpbSpinLock(OldIrql);
}

VOID
Ext2ClearVpbFlag (IN PVPB     Vpb,
		  IN USHORT   Flag )
{
	KIRQL OldIrql;
	
	IoAcquireVpbSpinLock(&OldIrql);
	
	Vpb->Flags &= ~Flag;
	
	IoReleaseVpbSpinLock(OldIrql);
}


NTSTATUS
Ext2BuildRequest (IN PDEVICE_OBJECT   DeviceObject,
		  IN PIRP             Irp)
{
	BOOLEAN             AtIrqlPassiveLevel = FALSE;
	BOOLEAN             IsTopLevelIrp = FALSE;
	PEXT2_IRP_CONTEXT   IrpContext = NULL;
	NTSTATUS            Status = STATUS_UNSUCCESSFUL;
	
	__try
	{
		__try
		{
//			Ext2DbgPrintCall(DeviceObject, Irp);
			
			AtIrqlPassiveLevel = (KeGetCurrentIrql() == PASSIVE_LEVEL);
			
			if (AtIrqlPassiveLevel)
			{
				FsRtlEnterFileSystem();
			}
			
			if (!IoGetTopLevelIrp())
			{
				IsTopLevelIrp = TRUE;
				IoSetTopLevelIrp(Irp);
			}
			
			IrpContext = Ext2AllocateIrpContext(DeviceObject, Irp);
			
			if (!IrpContext)
			{
				Status = STATUS_INSUFFICIENT_RESOURCES;
				Irp->IoStatus.Status = Status;
				Ext2CompleteRequest(Irp, IO_NO_INCREMENT);
			}
			else
			{
				Status = Ext2DispatchRequest(IrpContext);
			}
		}
		__except (Ext2ExceptionFilter(IrpContext, GetExceptionCode()))
		{
			Status = Ext2ExceptionHandler(IrpContext);
		}
	}
	__finally
	{
		if (IsTopLevelIrp)
		{
			IoSetTopLevelIrp(NULL);
		}
		
		if (AtIrqlPassiveLevel)
		{
			FsRtlExitFileSystem();
		}		
	}
	
	return Status;
}


VOID
DriverUnload (IN PDRIVER_OBJECT DriverObject)
/*
 * FUNCTION: Called by the system to unload the driver
 * ARGUMENTS:
 *           DriverObject = object describing this driver
 * RETURNS:  None
 */
{
#if DBG
	UNICODE_STRING	DosDeviceName;
#endif

	KdPrint(("Ext2Fsd: Unloading routine.\n"));
#if DBG
	RtlInitUnicodeString(&DosDeviceName, DOS_DEVICE_NAME);
	IoDeleteSymbolicLink(&DosDeviceName);
#endif
	ExDeleteResourceLite(&Ext2Global->Resource);
	
	IoDeleteDevice(Ext2Global->DeviceObject);

}

NTSTATUS
DriverEntry (
	     IN PDRIVER_OBJECT   DriverObject,
	     IN PUNICODE_STRING  RegistryPath
	     )
/*
 * FUNCTION: Called by the system to initalize the driver
 * ARGUMENTS:
 *           DriverObject = object describing this driver
 *           RegistryPath = path to our configuration entries
 * RETURNS: Success or failure
 */
{
	PDEVICE_OBJECT		    DeviceObject;
	PFAST_IO_DISPATCH           FastIoDispatch;
	PCACHE_MANAGER_CALLBACKS    CacheManagerCallbacks;
	PEXT2FS_EXT		    DeviceExt;

	UNICODE_STRING              DeviceName;
#if DBG
	UNICODE_STRING              DosDeviceName;
#endif
	NTSTATUS                    Status;

	LARGE_INTEGER		    CurrentTime;
	LARGE_INTEGER		    LocalTime;
	TIME_FIELDS				TimeFields;

	KeQuerySystemTime( &CurrentTime);
	KdPrint(("CurrentTime: %xh\n", CurrentTime.QuadPart));
	RtlTimeToTimeFields(
		&CurrentTime,
		&TimeFields);
	KdPrint(("CurrentTime: %d-%d-%d %d:%d:%d:%d\n",
		TimeFields.Year,
		TimeFields.Month,
		TimeFields.Day,
		TimeFields.Hour,
		TimeFields.Minute,
		TimeFields.Second,
		TimeFields.Milliseconds));

	RtlZeroMemory(&TimeFields, sizeof(TIME_FIELDS));
	TimeFields.Year = 1970;
	TimeFields.Month = 1;
	TimeFields.Day = 1;
	TimeFields.Hour = 8;
	TimeFields.Minute = 0;
	TimeFields.Second = 0;
	TimeFields.Milliseconds = 0;

	if (RtlTimeFieldsToTime(
		&TimeFields,
		&(LocalTime)))
	{
		KdPrint(("LocalTime: %I64xh\n", LocalTime.QuadPart));
	}
	else
		KdPrint(("FiledstoTime error.\n"));

  
	KdPrint(("Ext2 File System Driver by Matt Wu.\n"));
	KdPrint(("Ext2 DriverEntry...\n"));

	RtlInitUnicodeString(&DeviceName, DEVICE_NAME);
	
	Status = IoCreateDevice(
		DriverObject,
		sizeof(EXT2FS_EXT),
		&DeviceName,
		FILE_DEVICE_DISK_FILE_SYSTEM,
		0,
		FALSE,
		&DeviceObject );
	
	if (!NT_SUCCESS(Status))
	{
		KdPrint(("Ext2 IoCreateDevice error.\n"));
		return Status;
	}

	DeviceExt = (PEXT2FS_EXT) DeviceObject->DeviceExtension;
	RtlZeroMemory(DeviceExt, sizeof(EXT2FS_EXT));
	
	Ext2Global = &(DeviceExt->Ext2Global);
	Ext2Global->TimeZone.QuadPart = LocalTime.QuadPart;
	KdPrint(("TimeZone: %I64xh\n", Ext2Global->TimeZone.QuadPart));


	Ext2Global->Identifier.Type = FGD;
	Ext2Global->Identifier.Size = sizeof(EXT2_GLOBAL);
	Ext2Global->DeviceObject = DeviceObject;
	Ext2Global->DriverObject = DriverObject;
		
	DriverObject->MajorFunction[IRP_MJ_CREATE] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_READ] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_QUERY_INFORMATION] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_SET_INFORMATION] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_QUERY_VOLUME_INFORMATION] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_DIRECTORY_CONTROL] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_LOCK_CONTROL] = Ext2BuildRequest;
	DriverObject->MajorFunction[IRP_MJ_CLEANUP] = Ext2BuildRequest;
	DriverObject->DriverUnload = NULL;

	//
	// Initialize the fast I/O entry points
	//
	
	FastIoDispatch = &(Ext2Global->FastIoDispatch);
	
	FastIoDispatch->SizeOfFastIoDispatch = sizeof(FAST_IO_DISPATCH);
	FastIoDispatch->FastIoCheckIfPossible = Ext2FastIoCheckIfPossible;
#if DBG
	FastIoDispatch->FastIoRead = Ext2FastIoRead;
#else
	FastIoDispatch->FastIoRead = FsRtlCopyRead;
#endif
	FastIoDispatch->FastIoQueryBasicInfo = Ext2FastIoQueryBasicInfo;
	FastIoDispatch->FastIoQueryStandardInfo = Ext2FastIoQueryStandardInfo;
	FastIoDispatch->FastIoLock = Ext2FastIoLock;
	FastIoDispatch->FastIoUnlockSingle = Ext2FastIoUnlockSingle;
	FastIoDispatch->FastIoUnlockAll = Ext2FastIoUnlockAll;
	FastIoDispatch->FastIoUnlockAllByKey = Ext2FastIoUnlockAllByKey;
	FastIoDispatch->FastIoQueryNetworkOpenInfo = Ext2FastIoQueryNetworkOpenInfo;

	DriverObject->FastIoDispatch = FastIoDispatch;
	
	//
	// Initialize the Cache Manager callbacks
	//
	
	CacheManagerCallbacks = &(Ext2Global->CacheManagerCallbacks);
	CacheManagerCallbacks->AcquireForLazyWrite = Ext2AcquireForLazyWrite;
	CacheManagerCallbacks->ReleaseFromLazyWrite = Ext2ReleaseFromLazyWrite;
	CacheManagerCallbacks->AcquireForReadAhead = Ext2AcquireForReadAhead;
	CacheManagerCallbacks->ReleaseFromReadAhead = Ext2ReleaseFromReadAhead;

	//
	// Initialize the global data
	//

	InitializeListHead(&(Ext2Global->VcbList));
	ExInitializeResourceLite(&(Ext2Global->Resource));
#if DBG
	RtlInitUnicodeString(&DosDeviceName, DOS_DEVICE_NAME);
	IoCreateSymbolicLink(&DosDeviceName, &DeviceName);
#endif
	IoRegisterFileSystem(DeviceObject);
	
	return Status;
}
